from .project import *
from .definitions import *
from .classes import *
from .fields import *
from .records import *
from .assets import *
from .queries import *


class UnicatMutate:
    def __init__(self, unicat):
        self._unicat = unicat

        self.add_language = self._wrap(add_language)
        self.remove_language = self._wrap(remove_language)
        self.create_channel = self._wrap(create_channel)
        self.delete_channel = self._wrap(delete_channel)
        self.create_ordering = self._wrap(create_ordering)
        self.delete_ordering = self._wrap(delete_ordering)
        self.create_fieldlist = self._wrap(create_fieldlist)
        self.delete_fieldlist = self._wrap(delete_fieldlist)

        self.create_definition = self._wrap(create_definition)
        self.modify_definition = self._wrap(modify_definition)
        self.modify_definition_modify_layout = self._wrap(
            modify_definition_modify_layout
        )
        self.modify_definition_add_class = self._wrap(modify_definition_add_class)
        self.modify_definition_remove_class = self._wrap(modify_definition_remove_class)
        self.modify_definition_add_field = self._wrap(modify_definition_add_field)
        self.modify_definition_remove_field = self._wrap(modify_definition_remove_field)
        self.modify_definition_fieldlist_add_field = self._wrap(
            modify_definition_fieldlist_add_field
        )
        self.modify_definition_fieldlist_remove_field = self._wrap(
            modify_definition_fieldlist_remove_field
        )
        self.modify_definition_add_childdefinition = self._wrap(
            modify_definition_add_childdefinition
        )
        self.modify_definition_remove_childdefinition = self._wrap(
            modify_definition_remove_childdefinition
        )
        self.commit_definition = self._wrap(commit_definition)
        self.save_as_new_definition = self._wrap(save_as_new_definition)
        self.delete_definition = self._wrap(delete_definition)

        self.create_class = self._wrap(create_class)
        self.modify_class = self._wrap(modify_class)
        self.modify_class_modify_layout = self._wrap(modify_class_modify_layout)
        self.modify_class_add_field = self._wrap(modify_class_add_field)
        self.modify_class_remove_field = self._wrap(modify_class_remove_field)
        self.commit_class = self._wrap(commit_class)
        self.save_as_new_class = self._wrap(save_as_new_class)
        self.delete_class = self._wrap(delete_class)

        self.create_field = self._wrap(create_field)
        self.modify_field = self._wrap(modify_field)
        self.commit_field = self._wrap(commit_field)
        self.save_as_new_field = self._wrap(save_as_new_field)
        self.delete_field = self._wrap(delete_field)

        self.create_record = self._wrap(create_record)
        self.set_record_definition = self._wrap(set_record_definition)
        self.extend_record_definition_add_class = self._wrap(
            extend_record_definition_add_class
        )
        self.extend_record_definition_add_field = self._wrap(
            extend_record_definition_add_field
        )
        self.extend_record_definition_add_fieldlist_field = self._wrap(
            extend_record_definition_add_fieldlist_field
        )
        self.update_record = self._wrap(update_record)
        self.set_record_channels = self._wrap(set_record_channels)
        self.copy_record_channels_from_parent = self._wrap(
            copy_record_channels_from_parent
        )
        self.copy_record_channels_down = self._wrap(copy_record_channels_down)
        self.copy_record_channels_up = self._wrap(copy_record_channels_up)
        self.set_record_orderings = self._wrap(set_record_orderings)
        self.link_record = self._wrap(link_record)
        self.delete_record = self._wrap(delete_record)
        self.undelete_record = self._wrap(undelete_record)
        self.permanent_delete_record = self._wrap(permanent_delete_record)

        self.upload_asset = self._wrap(upload_asset)
        self.upload_update_asset = self._wrap(upload_update_asset)
        self.create_asset = self._wrap(create_asset)
        self.update_asset = self._wrap(update_asset)
        self.delete_asset = self._wrap(delete_asset)
        self.undelete_asset = self._wrap(undelete_asset)
        self.permanent_delete_asset = self._wrap(permanent_delete_asset)

        self.create_query = self._wrap(create_query)
        self.update_query = self._wrap(update_query)
        self.delete_query = self._wrap(delete_query)

        self.create_backup = self._wrap(create_backup)
        self.update_backup = self._wrap(update_backup)
        self.restore_backup = self._wrap(restore_backup)
        self.delete_backup = self._wrap(delete_backup)
        self.delete_backups = self._wrap(delete_backups)

    def _wrap(self, function):
        def _inner_function(*args, **kwargs):
            return function(self._unicat, *args, **kwargs)

        return _inner_function



import pytest
from unicat import Unicat
from mockapi import MockApi, MockApiFailure


unicat_url = "mocks://unicat.app"
project_gid = "<project-1>"
secret_api_key = "<secret_api_key>"
asset_folder = "/tmp/unicat"


@pytest.fixture
def unicat():
    unicat = Unicat(unicat_url, project_gid, secret_api_key, asset_folder)
    unicat.api = MockApi(unicat_url, project_gid, asset_folder)
    unicat.users = unicat.api.data["cc.users"]
    unicat.projects = unicat.api.data["cc.projects"]
    unicat.projects_members = unicat.api.data["cc.projects_members"]
    unicat.languages = unicat.api.data["cc.languages"]
    unicat.records = unicat.api.data["records"]
    unicat.definitions = unicat.api.data["definitions"]
    unicat.classes = unicat.api.data["classes"]
    unicat.fields = unicat.api.data["fields"]
    unicat.layouts = unicat.api.data["layouts"]
    unicat.assets = unicat.api.data["assets"]
    unicat.queries = unicat.api.data["queries"]
    return unicat


@pytest.fixture
def unicaterror():
    unicat = Unicat(unicat_url, project_gid, secret_api_key, asset_folder)
    unicat.api = MockApiFailure(unicat_url, project_gid, asset_folder)
    unicat.users = unicat.api.data["cc.users"]
    unicat.projects = unicat.api.data["cc.projects"]
    unicat.projects_members = unicat.api.data["cc.projects_members"]
    unicat.languages = unicat.api.data["cc.languages"]
    unicat.records = unicat.api.data["records"]
    unicat.definitions = unicat.api.data["definitions"]
    unicat.classes = unicat.api.data["classes"]
    unicat.fields = unicat.api.data["fields"]
    unicat.layouts = unicat.api.data["layouts"]
    unicat.assets = unicat.api.data["assets"]
    unicat.queries = unicat.api.data["queries"]
    return unicat



from .api import UnicatApi
from .error import UnicatError

from .project import UnicatProject
from .user import UnicatUser
from .record import UnicatRecord
from .asset import UnicatAsset

from .definition import UnicatDefinition
from .class_ import UnicatClass
from .field import UnicatField
from .query import UnicatQuery

from .mutate import UnicatMutate

from .transform import transformAsOptions

import json


class Unicat:
    """Higher-level Unicat API.

    Includes helpers for walking the tree, asset tree, and query results."""

    def __init__(self, unicat_url, project_gid, secret_api_key, asset_folder):

        self._unicat_url = unicat_url
        self._project_gid = project_gid
        self._secret_api_key = secret_api_key
        self.api = UnicatApi(self._unicat_url, self._project_gid, asset_folder)
        self.users = self.api.data["cc.users"]
        self.projects = self.api.data["cc.projects"]
        self.projects_members = self.api.data["cc.projects_members"]
        self.languages = self.api.data["cc.languages"]
        self.records = self.api.data["records"]
        self.definitions = self.api.data["definitions"]
        self.classes = self.api.data["classes"]
        self.fields = self.api.data["fields"]
        self.layouts = self.api.data["layouts"]
        self.assets = self.api.data["assets"]
        self.assets_by_pathname = self.api.data["assets_by_pathname"]
        self.queries = self.api.data["queries"]
        self._definitions_by_name = {}
        self._classes_by_name = {}
        self._fields_by_name = {}
        self._queries_by_type_name = {}
        self.mutate = UnicatMutate(self)

    def connect(self):
        success, result = self.api.connect(self._secret_api_key)
        return success

    @property
    def project(self):
        return UnicatProject(self, self._project_gid)

    def get_record(self, gid, *, force=False):
        if not gid:
            return None
        if gid not in self.records or force:
            success, result = self.api.call("/records/get", {"record": gid})
            if not success:
                raise UnicatError("get_record", result)
        record = UnicatRecord(self, gid)
        return record

    def get_records(self, gids, *, force=False):
        if not len(gids):
            return []
        if not self.records.keys() >= set(gids) or force:
            success, result = self.api.call("/records/get", {"records": gids})
            if not success:
                raise UnicatError("get_records", result)
            records = [UnicatRecord(self, gid) for gid in result["records"]]
        else:
            records = [UnicatRecord(self, gid) for gid in gids]
        return records

    def get_root_record(self):
        success, result = self.api.call("/records/root")
        if not success:
            raise UnicatError("get_root_record", result)
        root_record = UnicatRecord(self, result["root"])
        return root_record

    def get_asset(self, gid, *, force=False):
        if not gid:
            return None
        if gid not in self.assets or force:
            success, result = self.api.call("/assets/get", {"asset": gid})
            if not success:
                raise UnicatError("get_asset", result)
        asset = UnicatAsset(self, gid)
        return asset

    def get_asset_by_pathname(self, pathname, *, force=False):
        if not pathname:
            return None
        if pathname not in self.assets_by_pathname or force:
            success, result = self.api.call("/assets/get", {"pathname": pathname})
            if not success:
                raise UnicatError("get_asset_by_pathname", result)
            asset = UnicatAsset(self, result["asset"])
        else:
            asset = UnicatAsset(self, self.assets_by_pathname[pathname]["gid"])
        return asset

    def get_assets(self, gids, *, force=False):
        if not len(gids):
            return []
        if not self.assets.keys() >= set(gids) or force:
            success, result = self.api.call("/assets/get", {"assets": gids})
            if not success:
                raise UnicatError("get_assets", result)
            assets = [UnicatAsset(self, gid) for gid in result["assets"]]
        else:
            assets = [UnicatAsset(self, gid) for gid in gids]
        return assets

    def get_assets_by_pathnames(self, pathnames, *, force=False):
        if not len(pathnames):
            return []
        if not self.assets_by_pathname.keys() >= set(pathnames) or force:
            success, result = self.api.call("/assets/get", {"pathnames": pathnames})
            if not success:
                raise UnicatError("get_assets_by_pathnames", result)
            assets = [UnicatAsset(self, gid) for gid in result["assets"]]
        else:
            assets = [
                UnicatAsset(self, self.assets_by_pathname[pathname]["gid"])
                for pathname in pathnames
            ]
        return assets

    def get_root_asset(self):
        success, result = self.api.call("/assets/root")
        if not success:
            raise UnicatError("get_root_asset", result)
        root_asset = UnicatAsset(self, result["root"])
        return root_asset

    def get_definition(self, gid):
        if not gid:
            return None
        if gid not in self.definitions:
            return None
        definition = UnicatDefinition(self, gid)
        return definition

    def get_definitions(self, gids):
        if not len(gids):
            return []
        definitions = [
            UnicatDefinition(self, gid) for gid in gids if gid in self.definitions
        ]
        return definitions

    def _update_definitions_by_name(self):
        for definition in self.definitions.values():
            if not definition["is_new"] and not definition["is_working_copy"]:
                self._definitions_by_name[definition["name"]] = definition["gid"]

    def get_definition_by_name(self, name):
        if not name:
            return None
        if name not in self._definitions_by_name:
            self._update_definitions_by_name()
        if name not in self._definitions_by_name:
            return None
        definition = UnicatDefinition(self, self._definitions_by_name.get(name))
        return definition

    def get_definitions_by_name(self, names):
        if not len(names):
            return []
        if not self._definitions_by_name.keys() >= set(names):
            self._update_definitions_by_name()
        definitions_by_name = {
            name: UnicatDefinition(self, self._definitions_by_name.get(name))
            if name in self._definitions_by_name
            else None
            for name in names
        }
        return definitions_by_name

    def get_class(self, gid):
        if not gid:
            return None
        if gid not in self.classes:
            return None
        class_ = UnicatClass(self, gid)
        return class_

    def get_classes(self, gids):
        if not len(gids):
            return []
        classes = [UnicatClass(self, gid) for gid in gids if gid in self.classes]
        return classes

    def _update_classes_by_name(self):
        for class_ in self.classes.values():
            if not class_["is_new"] and not class_["is_working_copy"]:
                self._classes_by_name[class_["name"]] = class_["gid"]

    def get_class_by_name(self, name):
        if not name:
            return None
        if name not in self._classes_by_name:
            self._update_classes_by_name()
        if name not in self._classes_by_name:
            return None
        class_ = UnicatClass(self, self._classes_by_name.get(name))
        return class_

    def get_classes_by_name(self, names):
        if not len(names):
            return []
        if not self._classes_by_name.keys() >= set(names):
            self._update_classes_by_name()
        classes_by_name = {
            name: UnicatClass(self, self._classes_by_name.get(name))
            if name in self._classes_by_name
            else None
            for name in names
        }
        return classes_by_name

    def get_field(self, gid):
        if not gid:
            return None
        if gid not in self.fields:
            return None
        field = UnicatField(self, gid)
        return field

    def get_fields(self, gids):
        if not len(gids):
            return []
        fields = [UnicatField(self, gid) for gid in gids if gid in self.fields]
        return fields

    def _update_fields_by_name(self):
        for field in self.fields.values():
            if not field["is_new"] and not field["is_working_copy"]:
                self._fields_by_name[field["name"]] = field["gid"]

    def get_field_by_name(self, name):
        if not name:
            return None
        if name not in self._fields_by_name:
            self._update_fields_by_name()
        if name not in self._fields_by_name:
            return None
        field = UnicatField(self, self._fields_by_name.get(name))
        return field

    def get_fields_by_name(self, names):
        if not len(names):
            return []
        if not self._fields_by_name.keys() >= set(names):
            self._update_fields_by_name()
        fields_by_name = {
            name: UnicatField(self, self._fields_by_name.get(name))
            if name in self._fields_by_name
            else None
            for name in names
        }
        return fields_by_name

    def get_query(self, gid):
        if not gid:
            return None
        if gid not in self.queries:
            return None
        query = UnicatQuery(self, gid)
        return query

    def get_queries(self, gids):
        if not len(gids):
            return []
        queries = [UnicatQuery(self, gid) for gid in gids if gid in self.queries]
        return queries

    def _query_type_name_key(self, type, name):
        return type + ":" + name

    def _query_key(self, query):
        return self._query_type_name_key(query["type"], query["name"])

    def _update_queries_by_type_name(self):
        for query in self.queries.values():
            self._queries_by_type_name[self._query_key(query)] = query["gid"]

    def _get_query_by_type_name(self, type, name):
        if not name:
            return None
        key = self._query_type_name_key(type, name)
        if key not in self._queries_by_type_name:
            self._update_queries_by_type_name()
        if key not in self._queries_by_type_name:
            return None
        query = UnicatQuery(self, self._queries_by_type_name.get(key))
        return query

    def _get_queries_by_type_name(self, type, names):
        if not len(names):
            return []
        keys_by_name = {name: self._query_type_name_key(type, name) for name in names}
        if not self._queries_by_type_name.keys() >= set(keys_by_name.values()):
            self._update_queries_by_type_name()
        queries_by_name = {
            name: UnicatQuery(
                self,
                self._queries_by_type_name.get(self._query_type_name_key(type, name)),
            )
            if self._query_type_name_key(type, name) in self._queries_by_type_name
            else None
            for name in names
        }
        return queries_by_name

    def get_record_query_by_name(self, name):
        return self._get_query_by_type_name("record", name)

    def get_record_queries_by_name(self, names):
        return self._get_queries_by_type_name("record", names)

    def get_asset_query_by_name(self, name):
        return self._get_query_by_type_name("asset", name)

    def get_asset_queries_by_name(self, names):
        return self._get_queries_by_type_name("asset", names)

    def get_schema_query_by_name(self, name):
        return self._get_query_by_type_name("schema", name)

    def get_schema_queries_by_name(self, names):
        return self._get_queries_by_type_name("schema", names)

    def publish_asset(self, asset):
        if not asset.is_file:
            return None
        success, result = self.api.publish(asset._data)
        if not success:
            raise UnicatError("publish_asset", result)
        return result["public_url"]

    def publish_transformed_asset(self, asset, transform=None):
        if not asset.is_file:
            return None
        options = transformAsOptions(transform)
        success, result = self.api.transform(asset._data, options=options)
        if not success:
            raise UnicatError("publish_transformed_asset", result)
        return result["public_url"]

    def download_asset(self, asset, pathname=None, updated_on=None):
        if not asset.is_file:
            return None
        pathname = self.api.download(
            asset._data, pathname=pathname, updated_on=updated_on
        )
        return pathname

    def download_transformed_asset(
        self, asset, transform=None, pathname=None, updated_on=None
    ):
        if not asset.is_file:
            return None
        options = transformAsOptions(transform)
        pathname = self.api.download_transformed(
            asset._data, options=options, pathname=pathname, updated_on=updated_on
        )
        return pathname

    def walk_record_children(self, parent_record, channel=None, ordering=None):
        top = 0
        size = 100
        while True:
            data = {
                "record": parent_record.gid,
                "page.top": top,
                "page.size": size,
            }
            if channel is not None:
                data["channel"] = channel
            if ordering is not None:
                data["ordering"] = ordering
            success, result = self.api.call("/records/children", data)
            if not result["children"]:
                return
            for record_gid in result["children"]:
                record = UnicatRecord(self, record_gid)
                yield record
            top += size
            if top >= result["children.size"]:  # know when to stop
                break

    def walk_record_tree(self, channel=None, ordering=None):
        success, result = self.api.call("/records/root")
        root_record = UnicatRecord(self, result["root"])
        yield root_record
        yield from self._walk_record_tree_depth_first(
            root_record.gid, channel=channel, ordering=ordering
        )

    def _walk_record_tree_depth_first(self, parent_gid, channel=None, ordering=None):
        top = 0
        size = 100
        while True:
            data = {
                "record": parent_gid,
                "page.top": top,
                "page.size": size,
            }
            if channel is not None:
                data["channel"] = channel
            if ordering is not None:
                data["ordering"] = ordering
            success, result = self.api.call("/records/children", data)
            if not result["children"]:
                return
            for record_gid in result["children"]:
                record = UnicatRecord(self, record_gid)
                yield record
                if record.childcount > 0:
                    yield from self._walk_record_tree_depth_first(
                        record.gid, channel=channel, ordering=ordering
                    )
            top += size
            if top >= result["children.size"]:  # know when to stop
                break

    def walk_record_query(self, language, query, *, limit=None):
        request = {
            "selection": [
                {
                    "search": {
                        "language": language,
                        "q": query.q,
                        "filter": query.filter,
                    }
                },
            ],
            "page.size": min(100, limit if limit else 999),
        }
        result_count = 0
        while True:
            success, result = self.api.call("/records/bulk/get", request)
            if not success:
                raise UnicatError("walk_record_query", result)
            for record_gid in result["records"]:
                yield result["records.size"], UnicatRecord(self, record_gid)
                result_count += 1
                if limit and result_count >= limit:
                    return
            if not result["page.cursor"]:
                break
            request["page.cursor"] = result["page.cursor"]

    def walk_asset_children(self, parent_asset):
        top = 0
        size = 100
        while True:
            data = {
                "asset": parent_asset.gid,
                "page.top": top,
                "page.size": size,
            }
            success, result = self.api.call("/assets/children", data)
            if not result["children"]:
                return
            for asset_gid in result["children"]:
                asset = UnicatAsset(self, asset_gid)
                yield asset
            top += size
            if top >= result["children.size"]:  # know when to stop
                break

    def walk_asset_tree(self):
        success, result = self.api.call("/assets/root")
        root_asset = UnicatAsset(self, result["root"])
        yield root_asset
        yield from self._walk_asset_tree_depth_first(root_asset.gid)

    def _walk_asset_tree_depth_first(self, parent_gid):
        top = 0
        size = 100
        while True:
            data = {
                "asset": parent_gid,
                "page.top": top,
                "page.size": size,
            }
            success, result = self.api.call("/assets/children", data)
            if not result["children"]:
                return
            for asset_gid in result["children"]:
                asset = UnicatAsset(self, asset_gid)
                yield asset
                if asset.childcount > 0:
                    yield from self._walk_asset_tree_depth_first(asset.gid)
            top += size
            if top >= result["children.size"]:  # know when to stop
                break

    def walk_asset_query(self, language, query, *, limit=None):
        request = {
            "selection": [
                {
                    "search": {
                        "language": language,
                        "q": query.q,
                        "filter": query.filter,
                    }
                },
            ],
            "page.size": min(100, limit if limit else 999),
        }
        result_count = 0
        while True:
            success, result = self.api.call("/assets/bulk/get", request)
            if not success:
                raise UnicatError("walk_asset_query", result)
            for asset_gid in result["assets"]:
                yield result["assets.size"], UnicatAsset(self, asset_gid)
                result_count += 1
                if limit and result_count >= limit:
                    return
            if not result["page.cursor"]:
                break
            request["page.cursor"] = result["page.cursor"]

    def walk_queries(self):
        for gid in list(self.queries):
            yield UnicatQuery(self, gid)

    def walk_definitions(self):
        for gid in list(self.definitions):
            yield UnicatDefinition(self, gid)

    def walk_classes(self):
        for gid in list(self.classes):
            yield UnicatClass(self, gid)

    def walk_fields(self):
        for gid in list(self.fields):
            yield UnicatField(self, gid)

    def walk_schema_query(self, language, query, *, limit=None):
        request = {
            "selection": [
                {
                    "search": {
                        "language": language,
                        "q": query.q,
                        "filter": query.filter,
                    }
                },
            ],
            "page.size": min(100, limit if limit else 999),
        }
        UnicatSchemaItems = {
            "definition": UnicatDefinition,
            "class": UnicatClass,
            "field": UnicatField,
        }
        result_count = 0
        while True:
            success, result = self.api.call("/schema/bulk/get", request)
            if not success:
                raise UnicatError("walk_schema_query", result)
            for schema_item in result["schema_items"]:
                yield result["schema_items.size"], UnicatSchemaItem[
                    schema_item["type"]
                ](self, schema_item["item"])
                result_count += 1
                if limit and result_count >= limit:
                    return
            if not result["page.cursor"]:
                break
            request["page.cursor"] = result["page.cursor"]



import requests
import json
import os
from pathlib import Path
import logging
import time
from .utils import hash_data

logger = logging.getLogger("unicat.api")


class UnicatApi:
    """Connect to a Unicat project and have fun.

    Base usage:

        ccapi = UnicatApi(CONFIG["url"], CONFIG["project_gid"], CONFIG["local_asset_folder"])
        success, result = ccapi.connect(CONFIG["secret_api_key"])

        success, result = ccapi.call("/records/root", {"language": "en"})
        root_record = ccapi.data["records"][result["root"]]

    IMPORTANT:
        keep the API key secret, don't store it in code. See
        https://pypi.org/project/keyring/ for a safe way to use passwords and API keys.
    """

    def __init__(self, base_url, project_gid, asset_folder=None):
        self._requests_session = requests.Session()
        retries = requests.adapters.Retry(
            total=7, backoff_factor=0.1, allowed_methods=None
        )
        self._requests_session.mount(
            base_url, requests.adapters.HTTPAdapter(max_retries=retries)
        )

        self._base_url = base_url
        self._project_gid = project_gid
        self._asset_folder = asset_folder
        if self._asset_folder:
            os.makedirs(str(self._asset_folder), exist_ok=True)
        self._globalapi_url = base_url + "/api"
        self._api_url = base_url + "/api/p/" + project_gid
        self._sync_url = base_url + "/sync"
        self._dam_url = base_url + "/dam/p/" + project_gid
        self._auth_header = None
        self.cc_version = ""
        self.cc_cursor = 0
        self.project_gid = project_gid
        self.project_cursor = 0
        self.data = {
            "cc.users": {},
            "cc.projects": {},
            "cc.projects_members": {},
            "cc.languages": {},
            "records": {},
            "definitions": {},
            "classes": {},
            "fields": {},
            "layouts": {},
            "assets": {},
            "assets_by_pathname": {},
            "queries": {},
        }

    def _reset_data(self):
        self.data["cc.users"].clear()
        self.data["cc.projects"].clear()
        self.data["cc.projects_members"].clear()
        self.data["cc.languages"].clear()
        self.data["records"].clear()
        self.data["definitions"].clear()
        self.data["classes"].clear()
        self.data["fields"].clear()
        self.data["layouts"].clear()
        self.data["assets"].clear()
        self.data["assets_by_pathname"].clear()
        self.data["queries"].clear()

    def update_jwt(self, JWT):
        if not JWT:
            self._auth_header = None
        else:
            self._auth_header = {"Authorization": f"Bearer {JWT}"}

    def connect(self, secret_api_key):
        self._reset_data()
        data = {"api_key": secret_api_key}
        return self._call("post", "/auth", data)

    def init(self):
        self._reset_data()
        return self._call("get", "/init")

    def call(self, endpoint, data=None, method="post"):
        return self._call(method, endpoint, data=data)

    def upload(self, endpoint, data=None, localfilepath=None):
        return self._call_upload(endpoint, data=data, filepath=localfilepath)

    def publish(self, asset):
        if not asset["is_file"]:
            return False
        return self._call_dam("/publish", asset)

    def transform(self, asset, options=None):
        if not asset["is_file"]:
            return False
        if options and "key" not in options:
            options["key"] = hash_data(options)
        return self._call_dam("/transform", asset, options=options)

    def download(self, asset, pathname=None, updated_on=None):
        if not asset["is_file"]:
            return False
        if not self._asset_folder:
            return False
        success, result = self.publish(asset)
        if not success:
            return None
        if not pathname:
            path = os.path.dirname(asset["pathname"])
            name = os.path.basename(result["public_url"])
            pathname = path + "/" + name
        elif not pathname.startswith("/"):
            path = os.path.dirname(asset["pathname"])
            pathname = path + "/" + pathname
        return self._call_media(result["public_url"], pathname, updated_on=updated_on)

    def download_transformed(self, asset, options=None, pathname=None, updated_on=None):
        if not asset["is_file"]:
            return False
        if not self._asset_folder:
            return False
        success, result = self.transform(asset, options)
        if not success:
            return None
        if not pathname:
            path = os.path.dirname(asset["pathname"])
            name = os.path.basename(result["public_url"])
            pathname = path + "/" + name
        elif not pathname.startswith("/"):
            path = os.path.dirname(asset["pathname"])
            pathname = path + "/" + pathname
        return self._call_media(result["public_url"], pathname, updated_on=updated_on)

    def sync(self):
        data = {
            "cc_cursor": self.cc_cursor,
            "project": self._project_gid,
            "cursor": self.project_cursor,
        }
        success, result = self._call_sync("post", "/get", data=data)
        if not success:
            return None
        self.handle_sync_data(result["sync"])

    def prepare_for_job(self):
        data = {}
        success, result = self._call_sync("post", "/cursors", data=data)
        if not success:
            return None
        return result["cc.cursor"]

    def _new_job_sync_events(self, cursor):
        last_cursor = cursor
        while True:
            data = {"cc_cursor": last_cursor}
            success, result = self._call_sync("post", "/get", data=data)
            assert success
            for syncitem in result["sync"]:
                last_cursor = syncitem["cursor"]
                if syncitem["data_type"] == "jobs":
                    yield syncitem
            yield last_cursor

    def poll_job_sync_events(
        self, cursor, timeout_in_seconds=None, poll_interval_in_seconds=1.0
    ):
        used_time = 0
        new_job_sync_events_generator = self._new_job_sync_events(cursor)
        for syncitem in new_job_sync_events_generator:
            if type(syncitem) == int:  # cursor, means a single call was exhausted
                time.sleep(poll_interval_in_seconds)
                used_time += poll_interval_in_seconds
                if timeout_in_seconds and used_time >= timeout_in_seconds:
                    return
            else:
                yield syncitem

    def track_job(
        self, cursor, job_gid, timeout_in_seconds=None, poll_interval_in_seconds=1.0
    ):
        job_events = self.poll_job_sync_events(
            cursor,
            timeout_in_seconds=timeout_in_seconds,
            poll_interval_in_seconds=poll_interval_in_seconds,
        )
        for job_event in job_events:
            sync_cursor = job_event["cursor"]
            if job_event["data_key"] == job_gid:
                job_progress = {"gid": job_gid, **job_event["data"]}
                yield job_progress
                if job_progress["status"] == "done":
                    return

    def project_member_key(self, project_member):
        return f"{project_member['project_gid']}/{project_member['user_gid']}"

    def _fetch_syncdataitem(self, syncdataitem):
        type_ = syncdataitem["data_type"]
        key = syncdataitem["data_key"]
        if type_ == "cc.projects_members":
            project, member = key.split("/")
            success, result = self.call(  # note the // - global api
                "//members/get", {"project": project, "member": member}
            )
            return success
        map_calls = {
            "cc.users": ["//users/get", "user"],  # note the // - global api
            "cc.projects": [
                "//projects/get",
                "project",
            ],  # note the // - global api
            "assets": ["/assets/get", "asset"],
            "classes": ["/classes/get", "class"],
            "definitions": ["/definitions/get", "definition"],
            "fields": ["/fields/get", "field"],
            "layouts": ["/layouts/get", "layout"],
            "queries": ["/queries/get", "query"],
            "records": ["/records/get", "record"],
        }
        map_call = map_calls[type_]
        success, result = self.call(map_call[0], {map_call[1]: key})
        return success

    def handle_sync_data(self, syncdatalist):
        # result contains a list of cursor/action/data_type/data_key
        # handle each one, updating our cursors as we go
        for item in syncdatalist:
            # skip lagging syncs (older than our latest cursors)
            if item["data_type"] != "jobs":
                if item["type"] == "cc":
                    if self.cc_cursor >= item["cursor"]:
                        continue
                else:
                    if self.project_cursor >= item["cursor"]:
                        continue

            if item["data_type"] == "cc.version":
                if item["data_key"] != self.cc_version:
                    # alert! version-change mid-program!
                    print("Server version changed!")
                self.cc_version = item["data_key"]
            elif item["data_type"] == "jobs":
                job = item["data"]
                if job["job"] == "backup_project" and job["status"] == "queued":
                    print("Server database backup started")
                elif job["job"] == "backup_project" and job["status"] == "done":
                    print("Server database backup done")
                elif job["job"] == "restore_project" and job["status"] == "queued":
                    print("Server database restore started")
                elif job["job"] == "restore_project" and job["status"] == "done":
                    print("Server database restored done")
                    self.init()
            elif item["action"] == "DELETE":
                # use pop, not del, auto-handles items that aren't in our store
                self.data[item["data_type"]].pop(item["data_key"], None)
            elif item["action"] == "INSERT":
                # we're only interested in inserts that affect our data
                # so project-members for our project should fetch the new
                # membership, but also the new members
                # we're also interested in any base-data for definitions,
                # classes, fields, layouts, and queries
                # NOTE: fetching data auto-updates our local data-store
                if item["data_type"] == "cc.projects_members":
                    project_gid, user_gid = item["data_key"].split("/")
                    if project_gid == self._project_gid:
                        success = self._fetch_syncdataitem(item)
                elif item["data_type"] in (
                    "definitions",
                    "classes",
                    "fields",
                    "layouts",
                    "queries",
                ):
                    success = self._fetch_syncdataitem(item)
            elif item["action"] == "UPDATE":
                # we're only interested in data we already have locally
                if item["data_key"] in self.data[item["data_type"]]:
                    success = self._fetch_syncdataitem(item)
            # always update local cursors
            if item["type"] == "cc":
                self.cc_cursor = item["cursor"]
            else:
                self.project_cursor = item["cursor"]

    def _json_response(self, jsontext):
        try:
            jsondata = json.loads(jsontext)
        except:
            logger.info(f"500 Server error - Invalid response '''\n{jsontext}\n'''")
            return self._response_error(
                500, "Invalid response", info={"response.text": jsontext}
            )

        if "files" in jsondata["data"]:
            jsondata["result"]["data/files"] = jsondata["data"]["files"]
            del jsondata["data"]["files"]

        if jsondata["success"]:
            for key, value in jsondata["result"].items():
                if key == "cc.version":
                    self.cc_version = value
                elif key == "cc.cursor":
                    self.cc_cursor = value
                elif key == "cursor":
                    self.project_cursor = value
            for key, values in jsondata["data"].items():
                if key == "cc.projects_members":  # a list, not a dict
                    self.data[key].update(
                        {self.project_member_key(value): value for value in values}
                    )
                else:
                    self.data[key].update(values)
                if key == "assets":  # also reference by pathname
                    for asset in values.values():
                        self.data["assets_by_pathname"][asset["pathname"]] = self.data[
                            "assets"
                        ][asset["gid"]]
            return True, jsondata["result"]
        else:
            return False, jsondata["result"]

    def _response_error(self, code, message, info=None):
        return False, {"code": int(code), "message": message, "info": info}

    def _call(self, method, endpoint, data=None):
        if endpoint.startswith("//"):
            url = self._globalapi_url + endpoint[1:]
        else:
            url = self._api_url + endpoint
        requestmethod = getattr(self._requests_session, method)
        if method == "post" and data is None:
            data = {}
        response = requestmethod(url, headers=self._auth_header, json=data, timeout=60)
        logger.info(f"{response.status_code} {url}")
        logger.debug(
            f"{response.status_code} {url} REQUEST {json.dumps(data)} RESPONSE {response.text}"
        )
        if "Authorization" in response.headers:
            self.update_jwt(response.headers["Authorization"])
        if "WWW-Authenticate" in response.headers:
            self.update_jwt(None)
        jsontext = response.text
        return self._json_response(jsontext)

    def _call_sync(self, method, endpoint, data=None):
        url = self._sync_url + endpoint
        requestmethod = getattr(self._requests_session, method)
        if method == "post" and data is None:
            data = {}
        response = requestmethod(url, headers=self._auth_header, json=data, timeout=60)
        logger.info(f"{response.status_code} {url}")
        logger.debug(
            f"{response.status_code} {url} REQUEST {json.dumps(data)} RESPONSE {response.text}"
        )
        if "Authorization" in response.headers:
            self.update_jwt(response.headers["Authorization"])
        if "WWW-Authenticate" in response.headers:
            self.update_jwt(None)
        jsontext = response.text
        return self._json_response(jsontext)

    def _call_upload(self, endpoint, data, filepath):
        filesize = os.path.getsize(filepath)
        if filesize > 10_000_000:
            logger.info(f"413 Request Entity Too Large '{filepath}' ({filesize})")
            print()
            print(filepath, filesize)
            print()
            return self._response_error(413, "Request Entity Too Large", info=None)
        if not data:
            data = {}
        url = self._api_url + endpoint
        files = {"upload_file": (os.path.basename(filepath), open(filepath, "rb"))}
        response = self._requests_session.post(
            url, headers=self._auth_header, data=data, files=files, timeout=60
        )
        logger.info(f"{response.status_code} {url}")
        logger.debug(
            f"{response.status_code} {url} REQUEST {json.dumps(data)} FILE {files['upload_file'][0]} RESPONSE {response.text}"
        )
        if "Authorization" in response.headers:
            self.update_jwt(response.headers["Authorization"])
        if "WWW-Authenticate" in response.headers:
            self.update_jwt(None)
        jsontext = response.text
        return self._json_response(jsontext)

    def _call_dam(self, endpoint, asset, options=None):
        gid = asset["gid"]
        version = asset["version"]
        _, ext = os.path.splitext(asset["name"])
        url = self._dam_url + endpoint + f"/{gid}~{version}{ext}"
        if options:
            url += "/" + "/".join(
                f"{str(key)}={str(value)}" for key, value in options.items()
            )
        response = self._requests_session.get(
            url, headers=self._auth_header, timeout=60
        )
        logger.info(f"{response.status_code} {url}")
        logger.debug(
            f"{response.status_code} {url} REQUEST <get> RESPONSE {response.text}"
        )
        if "WWW-Authenticate" in response.headers:
            self.update_jwt(None)
        jsontext = response.text
        return self._json_response(jsontext)

    def _call_media(self, public_url, pathname, updated_on=None):
        if not self._asset_folder:
            return False
        while pathname.startswith("/") or pathname.startswith("\\"):
            pathname = pathname[1:]
        abs_pathname = Path(self._asset_folder, pathname)
        if os.path.isfile(abs_pathname):
            if updated_on is None:
                return pathname
            if updated_on <= os.path.getmtime(abs_pathname):
                return pathname
        with self._requests_session.get(
            public_url, stream=True, timeout=60
        ) as response:
            logger.info(f"{response.status_code} {public_url}")
            logger.debug(
                f"{response.status_code} {public_url} REQUEST <get> RESPONSE <bytes> INTO {abs_pathname}"
            )
            if response.status_code != 200:
                return False
            abs_path = os.path.dirname(abs_pathname)
            os.makedirs(abs_path, exist_ok=True)
            with open(abs_pathname, "wb") as f:
                for chunk in response.iter_content(chunk_size=8192):
                    f.write(chunk)
        return pathname



class UnicatQuery:
    def __init__(self, unicat, gid):
        self._unicat = unicat
        self._data = self._unicat.queries[gid]

    @property
    def gid(self):
        return self._data["gid"]

    @property
    def type(self):
        return self._data["type"]

    @property
    def name(self):
        return self._data["name"]

    @property
    def q(self):
        return self._data["q"]

    @property
    def filter(self):
        return self._data["filter"]



import pytest
from unicat import UnicatError, UnicatTransform
from setup import unicat, unicaterror


def test_connect(unicat):
    assert unicat.connect()


def test_connect_error(unicaterror):
    assert not unicaterror.connect()


def test_project_user_members(unicat):
    assert unicat.project.gid == "<project-1>"
    assert unicat.project.name == "Project the First"
    assert unicat.project.owner.name == "User the First"
    assert unicat.project.icon == "project-1-icon.jpg"
    assert unicat.project.status == "active"
    assert unicat.project.languages == ["en", "nl"]
    assert unicat.project.default_language == "en"
    assert unicat.project.channels["Channel the Second"] == "<channel-2>"
    assert unicat.project.default_channel == "<channel-1>"
    assert unicat.project.channel_name("<channel-1>") == "Channel the First"
    assert unicat.project.orderings["Ordering the First"] == "<ordering-1>"
    assert unicat.project.default_ordering == "<ordering-1>"
    assert unicat.project.ordering_name("<ordering-1>") == "Ordering the First"
    assert unicat.project.fieldlists["Fieldlist the First"] == "<fieldlist-1>"
    assert unicat.project.default_fieldlist == "<fieldlist-1>"
    assert unicat.project.fieldlist_name("<fieldlist-1>") == "Fieldlist the First"
    assert unicat.project.members[0].project.name == "Project the First"
    assert unicat.project.members[0].user.name == "User the First"
    owner = unicat.project.owner
    assert owner.gid == "<user-1>"
    assert owner.username == "user-1"
    assert owner.name == "User the First"
    assert owner.avatar == "user-1-avatar.png"
    member = unicat.project.members[0]
    assert member.key == "<project-1>/<user-1>"
    assert member.project.name == "Project the First"
    assert member.user.name == "User the First"
    assert member.status == "active"
    assert member.roles == [
        "owner",
    ]
    assert hasattr(member, "options")


def test_project_backups(unicat):
    assert unicat.project.get_backup(1).name == "Backup 1"
    assert unicat.project.get_backup(3).name == "Backup 3"
    assert len(unicat.project.backups) == 3
    assert unicat.project.backups[0].name == "Backup 3"
    assert unicat.project.backups[1].name == "Backup 2"
    assert unicat.project.backups[2].name == "Backup 1"
    assert unicat.project.get_backup(123) == None


def test_definition(unicat):
    language = unicat.project.default_language
    fieldlist = unicat.project.default_fieldlist
    definition = unicat.get_definition("<definition-1>")
    assert definition.gid == "<definition-1>"
    assert definition.original is None
    assert definition.name == "definition-1-name"
    assert definition.label[language] == "Definition 1"
    assert definition.classes[0].gid == "<class-1>"
    assert definition.classes_as_gids[0] == "<class-1>"
    assert definition.fields[0].gid == "<field-1>"
    assert definition.fields_as_gids[0] == "<field-1>"
    assert definition.titlefield.gid == "<field-1>"
    assert definition.fieldlists[fieldlist][0].gid == "<field-1>"
    assert definition.layout.gid == "<layout-1>"
    assert len(definition.childdefinitions) == 0
    assert definition.is_base == False
    assert definition.is_new == False
    assert definition.is_extended == False
    assert definition.is_working_copy == False
    assert definition.is_committed == True
    assert set(field.gid for field in definition.all_fields) == set(["<field-1>"])
    assert set(class_.gid for class_ in definition.base_classes) == set(["<class-1>"])
    assert set(field.gid for field in definition.base_fields) == set(["<field-1>"])
    assert set(field.gid for field in definition.all_base_fields) == set(["<field-1>"])
    assert definition.extended_classes == []
    assert definition.extended_fields == []
    assert definition.all_extended_fields == []


def test_class(unicat):
    language = unicat.project.default_language
    class_ = unicat.get_class("<class-1>")
    assert class_.gid == "<class-1>"
    assert class_.original is None
    assert class_.name == "class-1-name"
    assert class_.label[language] == "Class 1"
    assert class_.fields[0].gid == "<field-1>"
    assert class_.fields_as_gids[0] == "<field-1>"
    assert class_.layout.gid == "<layout-2>"
    assert class_.is_new == False
    assert class_.is_working_copy == False
    assert class_.is_committed == True


def test_field(unicat):
    language = unicat.project.default_language
    field = unicat.get_field("<field-1>")
    assert field.gid == "<field-1>"
    assert field.original is None
    assert field.name == "field-1-name"
    assert field.type == "text"
    assert field.class_ == None
    assert field.options == {}
    assert field.is_localized == True
    assert field.is_required == False
    assert field.label[language] == "Field 1"
    assert field.initial[language] == ""
    assert field.unit == "m²"
    assert field.title[language] == "Field 1 [m²]"
    assert field.is_new == False
    assert field.is_working_copy == False
    assert field.is_committed == True
    field2 = unicat.get_field("<field-2>")  # no unit
    assert field2.title[language] == "Field 2"


def test_layout(unicat):
    layout = unicat.get_definition("<definition-1>").layout
    assert layout.gid == "<layout-1>"
    assert layout.original is None
    assert layout.name == "layout-1-name"
    assert layout.root == "<component-1>"
    assert layout.components[layout.root]["type"] == "vertical"
    assert layout.is_new == False
    assert layout.is_working_copy == False


def test_query(unicat):
    query = unicat.get_query("<query-1>")
    assert query.gid == "<query-1>"
    assert query.type == "record"
    assert query.name == "query-1-record-name"
    assert query.q == ""
    assert query.filter == ["and", "", [["validation", "not_translated"]]]


def test_asset_folder(unicat):
    language = unicat.project.default_language
    transform = UnicatTransform(resize="fill", width=400, height=300)
    asset = unicat.get_asset("<asset-0>")
    assert asset.gid == "<asset-0>"
    assert asset.pathname == "/"
    assert asset.path == "/"
    assert asset.name == ""
    assert asset.is_file == False
    assert asset.type == None
    assert asset.childcount == 1
    assert asset.status == "published"
    assert asset.is_deleted == False
    assert "filecount" in asset.info
    assert asset.transforms is None
    assert asset.default_transform is None
    assert asset.title[language] == "Asset the Root"
    assert asset.description[language] == ""
    assert asset.updated_on == 1610635126.04762
    assert asset.publish() is None
    assert asset.publish_transformed(transform) is None
    assert asset.download() is None
    assert asset.download_transformed(transform) is None


def test_asset_file(unicat):
    language = unicat.project.default_language
    transform = UnicatTransform(resize="fill", width=400, height=300)
    asset = unicat.get_asset("<asset-1>")
    assert asset.gid == "<asset-1>"
    assert asset.pathname == "/asset-1-name.svg"
    assert asset.path == "/"
    assert asset.name == "asset-1-name.svg"
    assert asset.is_file == True
    assert asset.type == "svg"
    assert asset.childcount == 0
    assert asset.status == "published"
    assert asset.is_deleted == False
    assert "width" in asset.info
    assert "_main_" in asset.transforms
    assert "hotspot" in asset.default_transform
    assert asset.title[language] == "A generic image"
    assert (
        asset.description[language]
        == "Image - a line drawing of a frame containing mountains and the sun."
    )
    assert asset.updated_on == 1610635123.351925
    assert asset.publish() == "mocks://unicat.app/p/src/any-filename.ext"
    assert (
        asset.publish_transformed(transform)
        == "mocks://unicat.app/p/src/any-filename.ext"
    )
    assert asset.download() == "/tmp/unicat/any-filename.ext"
    assert asset.download_transformed(transform) == "/tmp/unicat/any-filename.ext"


def test_record(unicat):
    language = unicat.project.default_language
    ordering = unicat.project.default_ordering
    record = unicat.get_record("<record-1>")
    assert record.gid == "<record-1>"
    assert record.canonical == "<record-1>"  # should this be type UnicatRecord?
    assert record.parent == "<record-0>"  # should this be type UnicatRecord?
    assert record.backlinks == []  # should this be type list[UnicatRecord]?
    assert record.is_link == False
    assert record.is_deleted == False
    assert record.treelevel == 2
    assert record.path == ["<record-1>", "<record-0>"]
    assert record.title[language] == "Record the First"
    assert record.channels == ["__all__", "<channel-1>"]
    assert record.orderings[ordering] == 1
    assert record.childcount == 0
    assert record.definition.gid == "<definition-1>"
    assert type(record.fields[language]) is dict


def test_record_fields(unicat):
    language = unicat.project.default_language
    fields = unicat.get_record("<record-1>").fields[language]
    assert len(fields) == 1
    field = fields["field-1-name"]
    assert field.value == "Field 1 value 1"
    assert field.field.gid == "<field-1>"
    # TODO: make this work like field.gid == "<field-1>"
    #       and, for e.g. image fields, field.value.pathname == "/asset-1-name.svg"


def test_methods_reading(unicat):
    assert unicat.get_record("<record-1>").gid == "<record-1>"
    assert unicat.get_records(["<record-1>"])[0].gid == "<record-1>"
    assert unicat.get_root_record().gid == "<record-0>"
    assert unicat.get_asset("<asset-1>").gid == "<asset-1>"
    assert unicat.get_asset_by_pathname("asset-1-pathname").gid == "<asset-1>"
    assert unicat.get_assets(["<asset-1>"])[0].gid == "<asset-1>"
    assert unicat.get_root_asset().gid == "<asset-0>"
    assert unicat.get_definition("<definition-1>").gid == "<definition-1>"
    assert unicat.get_definitions(["<definition-1>"])[0].gid == "<definition-1>"
    assert unicat.get_definition_by_name("definition-1-name").gid == "<definition-1>"
    assert (
        unicat.get_definitions_by_name(["definition-1-name"])["definition-1-name"].gid
        == "<definition-1>"
    )
    assert unicat.get_class("<class-1>").gid == "<class-1>"
    assert unicat.get_classes(["<class-1>"])[0].gid == "<class-1>"
    assert unicat.get_class_by_name("class-1-name").gid == "<class-1>"
    assert (
        unicat.get_classes_by_name(["class-1-name"])["class-1-name"].gid == "<class-1>"
    )
    assert unicat.get_field("<field-1>").gid == "<field-1>"
    assert unicat.get_fields(["<field-1>"])[0].gid == "<field-1>"
    assert unicat.get_field_by_name("field-1-name").gid == "<field-1>"
    assert (
        unicat.get_fields_by_name(["field-1-name"])["field-1-name"].gid == "<field-1>"
    )
    assert unicat.get_query("<query-1>").gid == "<query-1>"
    assert unicat.get_queries(["<query-1>"])[0].gid == "<query-1>"
    assert unicat.get_record_query_by_name("query-1-record-name").gid == "<query-1>"
    assert (
        unicat.get_record_queries_by_name(["query-1-record-name"])[
            "query-1-record-name"
        ].gid
        == "<query-1>"
    )
    assert unicat.get_asset_query_by_name("query-2-asset-name").gid == "<query-2>"
    assert (
        unicat.get_asset_queries_by_name(["query-2-asset-name"])[
            "query-2-asset-name"
        ].gid
        == "<query-2>"
    )
    assert unicat.get_schema_query_by_name("query-3-schema-name").gid == "<query-3>"
    assert (
        unicat.get_schema_queries_by_name(["query-3-schema-name"])[
            "query-3-schema-name"
        ].gid
        == "<query-3>"
    )


def test_methods_reading_nonexisting(unicat):
    assert len(unicat.get_records(["<record-999>"])) == 0
    assert len(unicat.get_assets(["<asset-999>"])) == 0
    assert unicat.get_definition("<definition-999>") is None
    assert len(unicat.get_definitions(["<definition-999>"])) == 0
    assert unicat.get_definition_by_name("definition-999-name") is None
    assert (
        unicat.get_definitions_by_name(["definition-999-name"])["definition-999-name"]
        is None
    )
    assert unicat.get_class("<class-999>") is None
    assert len(unicat.get_classes(["<class-999>"])) == 0
    assert unicat.get_class_by_name("class-999-name") is None
    assert unicat.get_classes_by_name(["class-999-name"])["class-999-name"] is None
    assert unicat.get_field("<field-999>") is None
    assert len(unicat.get_fields(["<field-999>"])) == 0
    assert unicat.get_field_by_name("field-999-name") is None
    assert unicat.get_fields_by_name(["field-999-name"])["field-999-name"] is None
    assert unicat.get_query("<query-999>") is None
    assert len(unicat.get_queries(["<query-999>"])) == 0
    assert unicat.get_record_query_by_name("query-999-record-name") is None
    assert (
        unicat.get_record_queries_by_name(["query-999-record-name"])[
            "query-999-record-name"
        ]
        is None
    )
    assert unicat.get_asset_query_by_name("query-999-asset-name") is None
    assert (
        unicat.get_asset_queries_by_name(["query-999-asset-name"])[
            "query-999-asset-name"
        ]
        is None
    )
    assert unicat.get_schema_query_by_name("query-999-schema-name") is None
    assert (
        unicat.get_schema_queries_by_name(["query-999-schema-name"])[
            "query-999-schema-name"
        ]
        is None
    )


def test_methods_reading_error(unicaterror):
    with pytest.raises(UnicatError):
        unicaterror.get_record("<record-999>")
    with pytest.raises(UnicatError):
        unicaterror.get_records(["<record-999>"])
    with pytest.raises(UnicatError):
        unicaterror.get_root_record()
    with pytest.raises(UnicatError):
        unicaterror.get_asset("<asset-999>")
    with pytest.raises(UnicatError):
        unicaterror.get_asset_by_pathname("asset-999-pathname")
    with pytest.raises(UnicatError):
        unicaterror.get_assets(["<asset-999>"])
    with pytest.raises(UnicatError):
        unicaterror.get_root_asset()


def test_methods_traversing(unicat):
    record = unicat.get_record("<record-1>")
    asset = unicat.get_asset("<asset-1>")
    language = unicat.project.default_language
    record_query = unicat.get_query("<query-1>")
    asset_query = unicat.get_query("<query-2>")
    schema_query = unicat.get_query("<query-3>")
    assert hasattr(unicat.walk_record_children(record), "__iter__")
    assert hasattr(unicat.walk_record_tree(), "__iter__")
    assert hasattr(unicat.walk_record_query(language, record_query), "__iter__")
    assert hasattr(unicat.walk_asset_children(asset), "__iter__")
    assert hasattr(unicat.walk_asset_tree(), "__iter__")
    assert hasattr(unicat.walk_asset_query(language, asset_query), "__iter__")
    assert hasattr(unicat.walk_definitions(), "__iter__")
    assert hasattr(unicat.walk_classes(), "__iter__")
    assert hasattr(unicat.walk_fields(), "__iter__")
    assert hasattr(unicat.walk_schema_query(language, schema_query), "__iter__")
    assert hasattr(unicat.walk_queries(), "__iter__")



class UnicatClass:
    def __init__(self, unicat, gid):
        self._unicat = unicat
        self._data = self._unicat.classes[gid]

    @property
    def gid(self):
        return self._data["gid"]

    @property
    def original(self):
        if not self._data["original"]:
            return None
        return UnicatClass(self._unicat, self._data["original"])

    @property
    def name(self):
        return self._data["name"]

    @property
    def label(self):
        return self._data["label"]

    @property
    def fields(self):
        from .field import UnicatField

        return [UnicatField(self._unicat, gid) for gid in self._data["fields"]]

    @property
    def fields_as_gids(self):
        return self._data["fields"]

    @property
    def layout(self):
        from .layout import UnicatLayout

        return UnicatLayout(self._unicat, self._data["layout"])

    @property
    def is_new(self):
        return self._data["is_new"]

    @property
    def is_working_copy(self):
        return self._data["is_working_copy"]

    @property
    def is_committed(self):
        return not self._data["is_new"] and not self._data["is_working_copy"]



from ..error import UnicatError
from ..utils import _job


def add_language(unicat, language):
    success, result = unicat.api.call("/languages/add", {"language": language})
    if not success:
        raise UnicatError("add_language", result)
    return True


def remove_language(unicat, language):
    success, result = unicat.api.call("/languages/remove", {"language": language})
    if not success:
        raise UnicatError("remove_language", result)
    return True


def create_channel(unicat, name):
    success, result = unicat.api.call("/channels/create", {"name": name})
    if not success:
        raise UnicatError("create_channel", result)
    return result["channel"]


def delete_channel(unicat, channel_gid):
    success, result = unicat.api.call("/channels/delete", {"channel": channel_gid})
    if not success:
        raise UnicatError("delete_channel", result)
    return True


def create_ordering(unicat, name):
    success, result = unicat.api.call("/orderings/create", {"name": name})
    if not success:
        raise UnicatError("create_ordering", result)
    return result["ordering"]


def delete_ordering(unicat, ordering_gid):
    success, result = unicat.api.call("/orderings/delete", {"ordering": ordering_gid})
    if not success:
        raise UnicatError("delete_ordering", result)
    return True


def create_fieldlist(unicat, name):
    success, result = unicat.api.call("/fieldlists/create", {"name": name})
    if not success:
        raise UnicatError("create_fieldlist", result)
    return result["fieldlist"]


def delete_fieldlist(unicat, fieldlist_gid):
    success, result = unicat.api.call(
        "/fieldlists/delete", {"fieldlist": fieldlist_gid}
    )
    if not success:
        raise UnicatError("delete_fieldlist", result)
    return True


@_job
def create_backup(unicat, created_by, name=None):
    # becomes create_backup(unicat, created_by, name=None, *, return_job=False): UnicatProjectBackup | UnicatJob
    params = {"created_by": created_by}
    if name is not None:
        params["name"] = name
    success, result = unicat.api.call("/backups/create", params)
    if not success:
        raise UnicatError("create_backup", result)
    return unicat.project.get_backup(result["version"])


def update_backup(unicat, backup, name):
    success, result = unicat.api.call(
        "/backups/update", {"version": backup.version, "name": name}
    )
    if not success:
        raise UnicatError("update_backup", result)
    return unicat.project.get_backup(result["version"])


@_job
def restore_backup(unicat, backup):
    # becomes restore_backup(unicat, version, *, return_job=False): UnicatProject | UnicatJob
    success, result = unicat.api.call("/backups/restore", {"version": backup.version})
    if not success:
        raise UnicatError("restore_backup", result)
    return result["job.token"], unicat.project


def delete_backup(unicat, backup):
    success, result = unicat.api.call("/backups/delete", {"version": backup.version})
    if not success:
        raise UnicatError("delete_backup", result)
    return unicat.project


def delete_backups(unicat, backups):
    success, result = unicat.api.call(
        "/backups/delete", {"versions": [backup.version for backup in backups]}
    )
    if not success:
        raise UnicatError("delete_backups", result)
    return unicat.project



"""
Call loop machinery
"""
from __future__ import annotations

from typing import cast
from typing import Generator
from typing import Mapping
from typing import Sequence
from typing import Tuple
from typing import TYPE_CHECKING
from typing import Union

from ._result import _raise_wrapfail
from ._result import _Result
from ._result import HookCallError

if TYPE_CHECKING:
    from ._hooks import HookImpl


# Need to distinguish between old- and new-style hook wrappers.
# Wrapping one a singleton tuple is the fastest type-safe way I found to do it.
Teardown = Union[
    Tuple[Generator[None, _Result[object], None]],
    Generator[None, object, object],
]


def _multicall(
    hook_name: str,
    hook_impls: Sequence[HookImpl],
    caller_kwargs: Mapping[str, object],
    firstresult: bool,
) -> object | list[object]:
    """Execute a call into multiple python functions/methods and return the
    result(s).

    ``caller_kwargs`` comes from _HookCaller.__call__().
    """
    __tracebackhide__ = True
    results: list[object] = []
    exception = None
    only_new_style_wrappers = True
    try:  # run impl and wrapper setup functions in a loop
        teardowns: list[Teardown] = []
        try:
            for hook_impl in reversed(hook_impls):
                try:
                    args = [caller_kwargs[argname] for argname in hook_impl.argnames]
                except KeyError:
                    for argname in hook_impl.argnames:
                        if argname not in caller_kwargs:
                            raise HookCallError(
                                f"hook call must provide argument {argname!r}"
                            )

                if hook_impl.hookwrapper:
                    only_new_style_wrappers = False
                    try:
                        # If this cast is not valid, a type error is raised below,
                        # which is the desired response.
                        res = hook_impl.function(*args)
                        wrapper_gen = cast(Generator[None, _Result[object], None], res)
                        next(wrapper_gen)  # first yield
                        teardowns.append((wrapper_gen,))
                    except StopIteration:
                        _raise_wrapfail(wrapper_gen, "did not yield")
                elif hook_impl.wrapper:
                    try:
                        # If this cast is not valid, a type error is raised below,
                        # which is the desired response.
                        res = hook_impl.function(*args)
                        function_gen = cast(Generator[None, object, object], res)
                        next(function_gen)  # first yield
                        teardowns.append(function_gen)
                    except StopIteration:
                        _raise_wrapfail(function_gen, "did not yield")
                else:
                    res = hook_impl.function(*args)
                    if res is not None:
                        results.append(res)
                        if firstresult:  # halt further impl calls
                            break
        except BaseException as exc:
            exception = exc
    finally:
        # Fast path - only new-style wrappers, no _Result.
        if only_new_style_wrappers:
            if firstresult:  # first result hooks return a single value
                result = results[0] if results else None
            else:
                result = results

            # run all wrapper post-yield blocks
            for teardown in reversed(teardowns):
                try:
                    if exception is not None:
                        teardown.throw(exception)  # type: ignore[union-attr]
                    else:
                        teardown.send(result)  # type: ignore[union-attr]
                    # Following is unreachable for a well behaved hook wrapper.
                    # Try to force finalizers otherwise postponed till GC action.
                    # Note: close() may raise if generator handles GeneratorExit.
                    teardown.close()  # type: ignore[union-attr]
                except StopIteration as si:
                    result = si.value
                    exception = None
                    continue
                except BaseException as e:
                    exception = e
                    continue
                _raise_wrapfail(teardown, "has second yield")  # type: ignore[arg-type]

            if exception is not None:
                raise exception.with_traceback(exception.__traceback__)
            else:
                return result

        # Slow path - need to support old-style wrappers.
        else:
            if firstresult:  # first result hooks return a single value
                outcome: _Result[object | list[object]] = _Result(
                    results[0] if results else None, exception
                )
            else:
                outcome = _Result(results, exception)

            # run all wrapper post-yield blocks
            for teardown in reversed(teardowns):
                if isinstance(teardown, tuple):
                    try:
                        teardown[0].send(outcome)
                        _raise_wrapfail(teardown[0], "has second yield")
                    except StopIteration:
                        pass
                else:
                    try:
                        if outcome._exception is not None:
                            teardown.throw(outcome._exception)
                        else:
                            teardown.send(outcome._result)
                        # Following is unreachable for a well behaved hook wrapper.
                        # Try to force finalizers otherwise postponed till GC action.
                        # Note: close() may raise if generator handles GeneratorExit.
                        teardown.close()
                    except StopIteration as si:
                        outcome.force_result(si.value)
                        continue
                    except BaseException as e:
                        outcome.force_exception(e)
                        continue
                    _raise_wrapfail(teardown, "has second yield")

            return outcome.get_result()



class UnicatProjectMember:
    def __init__(self, unicat, project_member_data):
        self._unicat = unicat
        self._data = project_member_data

    @property
    def project(self):
        from .project import UnicatProject

        return UnicatProject(self._unicat, self._data["project_gid"])

    @property
    def user(self):
        from .user import UnicatUser

        return UnicatUser(self._unicat, self._data["user_gid"])

    @property
    def status(self):
        return self._data["status"]

    @property
    def roles(self):
        return self._data["roles"]

    @property
    def options(self):
        return self._data["options"]

    @property
    def key(self):
        return self._unicat.api.project_member_key(self._data)



import itertools

from .compat import collections_abc


class DirectedGraph(object):
    """A graph structure with directed edges."""

    def __init__(self):
        self._vertices = set()
        self._forwards = {}  # <key> -> Set[<key>]
        self._backwards = {}  # <key> -> Set[<key>]

    def __iter__(self):
        return iter(self._vertices)

    def __len__(self):
        return len(self._vertices)

    def __contains__(self, key):
        return key in self._vertices

    def copy(self):
        """Return a shallow copy of this graph."""
        other = DirectedGraph()
        other._vertices = set(self._vertices)
        other._forwards = {k: set(v) for k, v in self._forwards.items()}
        other._backwards = {k: set(v) for k, v in self._backwards.items()}
        return other

    def add(self, key):
        """Add a new vertex to the graph."""
        if key in self._vertices:
            raise ValueError("vertex exists")
        self._vertices.add(key)
        self._forwards[key] = set()
        self._backwards[key] = set()

    def remove(self, key):
        """Remove a vertex from the graph, disconnecting all edges from/to it."""
        self._vertices.remove(key)
        for f in self._forwards.pop(key):
            self._backwards[f].remove(key)
        for t in self._backwards.pop(key):
            self._forwards[t].remove(key)

    def connected(self, f, t):
        return f in self._backwards[t] and t in self._forwards[f]

    def connect(self, f, t):
        """Connect two existing vertices.

        Nothing happens if the vertices are already connected.
        """
        if t not in self._vertices:
            raise KeyError(t)
        self._forwards[f].add(t)
        self._backwards[t].add(f)

    def iter_edges(self):
        for f, children in self._forwards.items():
            for t in children:
                yield f, t

    def iter_children(self, key):
        return iter(self._forwards[key])

    def iter_parents(self, key):
        return iter(self._backwards[key])


class IteratorMapping(collections_abc.Mapping):
    def __init__(self, mapping, accessor, appends=None):
        self._mapping = mapping
        self._accessor = accessor
        self._appends = appends or {}

    def __repr__(self):
        return "IteratorMapping({!r}, {!r}, {!r})".format(
            self._mapping,
            self._accessor,
            self._appends,
        )

    def __bool__(self):
        return bool(self._mapping or self._appends)

    __nonzero__ = __bool__  # XXX: Python 2.

    def __contains__(self, key):
        return key in self._mapping or key in self._appends

    def __getitem__(self, k):
        try:
            v = self._mapping[k]
        except KeyError:
            return iter(self._appends[k])
        return itertools.chain(self._accessor(v), self._appends.get(k, ()))

    def __iter__(self):
        more = (k for k in self._appends if k not in self._mapping)
        return itertools.chain(self._mapping, more)

    def __len__(self):
        more = sum(1 for k in self._appends if k not in self._mapping)
        return len(self._mapping) + more


class _FactoryIterableView(object):
    """Wrap an iterator factory returned by `find_matches()`.

    Calling `iter()` on this class would invoke the underlying iterator
    factory, making it a "collection with ordering" that can be iterated
    through multiple times, but lacks random access methods presented in
    built-in Python sequence types.
    """

    def __init__(self, factory):
        self._factory = factory
        self._iterable = None

    def __repr__(self):
        return "{}({})".format(type(self).__name__, list(self))

    def __bool__(self):
        try:
            next(iter(self))
        except StopIteration:
            return False
        return True

    __nonzero__ = __bool__  # XXX: Python 2.

    def __iter__(self):
        iterable = (
            self._factory() if self._iterable is None else self._iterable
        )
        self._iterable, current = itertools.tee(iterable)
        return current


class _SequenceIterableView(object):
    """Wrap an iterable returned by find_matches().

    This is essentially just a proxy to the underlying sequence that provides
    the same interface as `_FactoryIterableView`.
    """

    def __init__(self, sequence):
        self._sequence = sequence

    def __repr__(self):
        return "{}({})".format(type(self).__name__, self._sequence)

    def __bool__(self):
        return bool(self._sequence)

    __nonzero__ = __bool__  # XXX: Python 2.

    def __iter__(self):
        return iter(self._sequence)


def build_iter_view(matches):
    """Build an iterable view from the value returned by `find_matches()`."""
    if callable(matches):
        return _FactoryIterableView(matches)
    if not isinstance(matches, collections_abc.Sequence):
        matches = list(matches)
    return _SequenceIterableView(matches)



from ..query import UnicatQuery
from ..error import UnicatError

typecheck = type  # I use 'type' as a keyword argument


def create_query(unicat, type, *, name, q=None, filter=None):
    properties = {}
    if type is not None and type in ("record", "asset", "schema"):
        properties["type"] = type
    if name is not None:
        properties["name"] = name
    if q is not None:
        properties["q"] = q
    if filter is not None and typecheck(filter) in (list, tuple):
        properties["filter"] = filter
    success, result = unicat.api.call("/queries/create", properties)
    if not success:
        raise UnicatError("create_query", result)
    return UnicatQuery(unicat, result["query"])


def update_query(unicat, query, *, name=None, q=None, filter=None):
    properties = {}
    if name is not None:
        properties["name"] = name
    if q is not None:
        properties["q"] = q
    if filter is not None and typecheck(filter) in (list, tuple):
        properties["filter"] = filter
    properties["query"] = query.gid
    success, result = unicat.api.call("/queries/update", properties)
    if not success:
        raise UnicatError("update_query", result)
    return UnicatQuery(unicat, result["query"])


def delete_query(unicat, query):
    success, result = unicat.api.call("/queries/delete", {"query": query.gid})
    if not success:
        raise UnicatError("delete_query", result)
    if query.gid in unicat.api.data["queries"]:
        del unicat.api.data["queries"][query.gid]
    return True



from ..definition import UnicatDefinition
from ..error import UnicatError


def create_definition(
    unicat,
    *,
    name=None,
    label=None,
    classes=None,
    fields=None,
    titlefield=None,
    childdefinitions=None,
    **kwargs,
):
    properties = {**kwargs}
    if name is not None:
        properties["name"] = name
    if label is not None and type(label) is dict:
        properties["label"] = label
    if classes is not None and type(classes) is not str:
        properties["classes"] = [
            class_.gid if hasattr(class_, "gid") else class_ for class_ in classes
        ]
    if fields is not None and type(fields) is not str:
        properties["fields"] = [
            field.gid if hasattr(field, "gid") else field for field in fields
        ]
    if titlefield is not None:
        properties["titlefield"] = (
            titlefield.gid if hasattr(titlefield, "gid") else titlefield
        )
    if childdefinitions is not None and type(childdefinitions) is not str:
        properties["childdefinitions"] = [
            childdefinition.gid if hasattr(childdefinition, "gid") else childdefinition
            for childdefinition in childdefinitions
        ]
    success, result = unicat.api.call("/definitions/create", properties)
    if not success:
        raise UnicatError("create_definition", result)
    return UnicatDefinition(unicat, result["definition"])


def modify_definition(
    unicat,
    definition,
    *,
    name=None,
    label=None,
    classes=None,
    fields=None,
    titlefield=None,
    childdefinitions=None,
    **kwargs,
):
    properties = {**kwargs}
    if name is not None:
        properties["name"] = name
    if label is not None and type(label) is dict:
        properties["label"] = label
    if classes is not None and type(classes) is not str:
        properties["classes"] = [
            class_.gid if hasattr(class_, "gid") else class_ for class_ in classes
        ]
    if fields is not None and type(fields) is not str:
        properties["fields"] = [
            field.gid if hasattr(field, "gid") else field for field in fields
        ]
    if titlefield is not None:
        properties["titlefield"] = (
            titlefield.gid if hasattr(titlefield, "gid") else titlefield
        )
    if childdefinitions is not None and type(childdefinitions) is not str:
        properties["childdefinitions"] = [
            childdefinition.gid if hasattr(childdefinition, "gid") else childdefinition
            for childdefinition in childdefinitions
        ]
    properties["definition"] = definition.gid
    success, result = unicat.api.call("/definitions/modify", properties)
    if not success:
        raise UnicatError("modify_definition", result)
    return UnicatDefinition(unicat, result["definition"])


def modify_definition_modify_layout(
    unicat,
    definition,
    *,
    name=None,
    root=None,
    components=None,
):
    layout_properties = {}
    if name is not None:
        layout_properties["name"] = name
    if root is not None:
        layout_properties["root"] = root
    if components is not None and type(components) is dict:
        layout_properties["components"] = components
    success, result = unicat.api.call(
        "/definitions/layouts/modify",
        {**layout_properties, "definition": definition.gid},
    )
    if not success:
        raise UnicatError("modify_definition_modify_layout", result)
    return UnicatDefinition(unicat, result["definition"])


def modify_definition_add_class(unicat, definition, class_):
    success, result = unicat.api.call(
        "/definitions/classes/add",
        {"definition": definition.gid, "class": class_.gid},
    )
    if not success:
        raise UnicatError("modify_definition_add_class", result)
    return UnicatDefinition(unicat, result["definition"])


def modify_definition_remove_class(unicat, definition, class_):
    success, result = unicat.api.call(
        "/definitions/classes/remove",
        {"definition": definition.gid, "class": class_.gid},
    )
    if not success:
        raise UnicatError("modify_definition_remove_class", result)
    return UnicatDefinition(unicat, result["definition"])


def modify_definition_add_field(unicat, definition, field):
    success, result = unicat.api.call(
        "/definitions/fields/add",
        {"definition": definition.gid, "field": field.gid},
    )
    if not success:
        raise UnicatError("modify_definition_add_field", result)
    return UnicatDefinition(unicat, result["definition"])


def modify_definition_remove_field(unicat, definition, field):
    success, result = unicat.api.call(
        "/definitions/fields/remove",
        {"definition": definition.gid, "field": field.gid},
    )
    if not success:
        raise UnicatError("modify_definition_remove_field", result)
    return UnicatDefinition(unicat, result["definition"])


def modify_definition_fieldlist_add_field(unicat, definition, fieldlist_key, field):
    success, result = unicat.api.call(
        "/definitions/fieldlists/fields/add",
        {
            "definition": definition.gid,
            "fieldlist_key": fieldlist_key,
            "field": field.gid,
        },
    )
    if not success:
        raise UnicatError("modify_definition_fieldlist_add_field", result)
    return UnicatDefinition(unicat, result["definition"])


def modify_definition_fieldlist_remove_field(unicat, definition, fieldlist_key, field):
    success, result = unicat.api.call(
        "/definitions/fieldlists/fields/remove",
        {
            "definition": definition.gid,
            "fieldlist_key": fieldlist_key,
            "field": field.gid,
        },
    )
    if not success:
        raise UnicatError("modify_definition_fieldlist_remove_field", result)
    return UnicatDefinition(unicat, result["definition"])


def modify_definition_add_childdefinition(unicat, definition, childdefinition):
    success, result = unicat.api.call(
        "/definitions/childdefinitions/add",
        {"definition": definition.gid, "childdefinition": childdefinition.gid},
    )
    if not success:
        raise UnicatError("modify_definition_add_childdefinition", result)
    return UnicatDefinition(unicat, result["definition"])


def modify_definition_remove_childdefinition(unicat, definition, childdefinition):
    success, result = unicat.api.call(
        "/definitions/childdefinitions/remove",
        {"definition": definition.gid, "childdefinition": childdefinition.gid},
    )
    if not success:
        raise UnicatError("modify_definition_remove_childdefinition", result)
    return UnicatDefinition(unicat, result["definition"])


def commit_definition(unicat, new_or_working_copy):
    success, result = unicat.api.call(
        "/definitions/commit", {"definition": new_or_working_copy.gid}
    )
    if not success:
        raise UnicatError("commit_definition", result)
    if (
        result["definition"] != new_or_working_copy.gid
        and new_or_working_copy.gid in unicat.api.data["definitions"]
    ):
        del unicat.api.data["definitions"][new_or_working_copy.gid]
    return UnicatDefinition(unicat, result["definition"])


def save_as_new_definition(unicat, working_copy):
    success, result = unicat.api.call(
        "/definitions/save_as_new", {"definition": working_copy.gid}
    )
    if not success:
        raise UnicatError("save_as_new_definition", result)
    return UnicatDefinition(unicat, result["definition"])


def delete_definition(unicat, definition):
    success, result = unicat.api.call(
        "/definitions/delete", {"definition": definition.gid}
    )
    if not success:
        raise UnicatError("delete_definition", result)
    if definition.gid in unicat.api.data["definitions"]:
        del unicat.api.data["definitions"][definition.gid]
    return True



import re

rgb_re = re.compile("^#?[0-9a-fA-F]{6}$")


def mergeTransforms(*transforms):
    """Return new transform by merging all given transforms"""
    mergedTransform = UnicatTransform()
    for transform in transforms:
        mergedTransform.merge(transform)
    return mergedTransform


def transformAsOptions(transform):
    if not isinstance(transform, UnicatTransform):
        return None
    options = {}
    if transform.name is not None:
        options["name"] = transform.name
    if transform.key is not None:
        options["key"] = transform.key
    if transform.type is not None:
        options["type"] = transform.type
    if transform.force is not None:
        options["force"] = "true" if transform.force else "false"
    if transform.optimize is not None:
        options["optimize"] = "true" if transform.optimize else "false"
    if transform.resize is not None:
        options[transform.resize] = f"{transform.width},{transform.height}"
        if transform.resize == "width" and not transform.height:
            options[transform.resize] = f"{transform.width}"
        elif transform.resize == "height" and not transform.width:
            options[transform.resize] = f"{transform.height}"
    if transform.hotspot is not None:
        options["hs"] = f"{transform.hotspot[0]:0.5f},{transform.hotspot[1]:0.5f}"
    if transform.crop is not None:
        options["crop"] = f"{transform.crop[0]:0.5f},{transform.crop[1]:0.5f}"
    if transform.padding is not None:
        options[
            "padding"
        ] = f"{transform.padding[0]},{transform.padding[1]},{transform.padding[2]},{transform.padding[3]}"
    if transform.background is not None:
        options["bg"] = transform.background
    if transform.quality is not None:
        options["q"] = f"{transform.quality}"
    if transform.dpr is not None:
        options["dpr"] = f"{transform.dpr}"
    return options


class UnicatTransform:
    name = None
    key = None
    force = None
    optimize = None
    resize = None
    width = None
    height = None
    type = None
    hotspot = None
    crop = None
    padding = None
    quality = None
    background = None
    dpr = None

    def __init__(self, **kwargs):
        setters = {
            "name": self.setName,
            "key": self.setKey,
            "force": self.setForce,
            "optimize": self.setOptimize,
            "resize": self.setResize,
            "width": self.setWidth,
            "height": self.setHeight,
            "type": self.setType,
            "hotspot": self.setHotspot,
            "crop": self.setCrop,
            "padding": self.setPadding,
            "quality": self.setQuality,
            "background": self.setBackground,
            "dpr": self.setDevicePixelRatio,
        }
        for name, value in kwargs.items():
            if name in setters:
                if type(value) is list or type(value) is tuple:
                    setters[name](*value)
                else:
                    setters[name](value)

    def setName(self, name):
        self.name = str(name)
        return self

    def setKey(self, key):
        self.key = str(key)
        return self

    def setForce(self, force):
        if not isinstance(force, bool):
            self.force = None
            return self
        self.force = force
        return self

    def setOptimize(self, optimize):
        if not isinstance(optimize, bool):
            self.optimize = None
            return self
        self.optimize = optimize
        return self

    def setResize(self, resize):
        if resize in ["width", "height", "fit", "fill"]:
            self.resize = resize
        else:
            self.resize = "width"
        # to make sure we have valid values
        self.setWidth(self.width)
        self.setHeight(self.height)
        return self

    def setWidth(self, width):
        if not width:
            if self.resize == "height":
                self.width = None
            else:
                self.width = 1
        else:
            self.width = max(1, min(round(width), 5000))
        return self

    def setHeight(self, height):
        if not height:
            if self.resize == "width":
                self.height = None
            else:
                self.height = 1
        else:
            self.height = max(1, min(round(height), 5000))
        return self

    def setType(self, type):
        if type in ["jpg", "png", "gif"]:
            self.type = type
        else:
            self.type = "jpg"
        return self

    def setHotspot(self, x, y):
        if not isinstance(x, (int, float)) or not isinstance(y, (int, float)):
            self.hotspot = None
            return self
        x = max(0.0, min(x, 1.0))
        y = max(0.0, min(y, 1.0))
        self.hotspot = [x, y]
        return self

    def setCrop(self, w, h):
        if not isinstance(w, (int, float)) or not isinstance(h, (int, float)):
            self.crop = None
            return self
        w = max(0.0, min(w, 1.0))
        h = max(0.0, min(h, 1.0))
        floatingpointthreshold = 0.00001
        if w < floatingpointthreshold or h < floatingpointthreshold:
            self.crop = None
        else:
            self.crop = [w, h]
            if not self.hotspot:
                self.setHotspot(0.5, 0.5)
        return self

    def setPadding(self, top, right, bottom, left):
        for pad in (top, right, bottom, left):
            if not (pad == "auto" or isinstance(pad, (int, float))):
                self.padding = None
                return self
        if top != "auto":
            top = max(0, min(top, 1000))
        if right != "auto":
            right = max(0, min(right, 1000))
        if bottom != "auto":
            bottom = max(0, min(bottom, 1000))
        if left != "auto":
            left = max(0, min(left, 1000))
        self.padding = [top, right, bottom, left]
        return self

    def setQuality(self, q):
        if not isinstance(q, (int, float)):
            self.quality = None
            return self
        q = max(0, min(round(q), 100))
        self.quality = q
        return self

    def setBackground(self, background):
        if background in ["white", "transparent"]:
            self.background = background
        else:
            if bool(rgb_re.match(background)):
                self.background = background.replace("#", "")
        return self

    def setDevicePixelRatio(self, dpr):
        if not isinstance(dpr, (int, float)):
            self.dpr = None
            return self
        dpr = max(1, min(round(dpr), 4))
        self.dpr = dpr
        return self

    def merge(self, transform):
        if not isinstance(transform, UnicatTransform):
            return self
        if transform.name is not None:
            self.setName(transform.name)
        if transform.key is not None:
            self.setKey(transform.key)
        if transform.force is not None:
            self.setForce(transform.force)
        if transform.optimize is not None:
            self.setOptimize(transform.optimize)
        if transform.resize is not None:
            self.setResize(transform.resize)
        if transform.width is not None:
            self.setWidth(transform.width)
        if transform.height is not None:
            self.setHeight(transform.height)
        if transform.type is not None:
            self.setType(transform.type)
        if transform.hotspot is not None:
            self.setHotspot(transform.hotspot[0], transform.hotspot[1])
        if transform.crop is not None:
            self.setCrop(transform.crop[0], transform.crop[1])
        if transform.padding is not None:
            self.setPadding(
                transform.padding[0],
                transform.padding[1],
                transform.padding[2],
                transform.padding[3],
            )
        if transform.quality is not None:
            self.setQuality(transform.quality)
        if transform.background is not None:
            self.setBackground(transform.background)
        if transform.dpr is not None:
            self.setDevicePixelRatio(transform.dpr)
        return self



import pytest
from unicat.utils import (
    gid,
    maybe_gid,
    test_false as alias_test_false,  # alias for pytest, can't start with test_
    test_true as alias_test_true,  # alias for pytest, can't start with test_
    DuckObject,
    convert_value_to_fielddata,
    convert_fielddata_to_value,
    diff_record_fields_data,
    merge_dicts,
    hash_text,
    hash_data,
)
from setup import unicat, unicaterror


def test_gid():
    key = gid()
    assert type(key) is str
    assert maybe_gid(key)
    assert not maybe_gid(key[1:])
    assert not maybe_gid(key[:-1])
    assert not maybe_gid(key + "1")
    assert not maybe_gid("1" + key)
    assert not maybe_gid(123)


def test_true_false():
    assert alias_test_false(False)
    assert alias_test_true(True)
    for false in (
        "",
        "0",
        "no",
        "n",
        "off",
        "false",
        "f",
        "unchecked",
        "deselected",
        "none",
        "null",
        "undefined",
    ):
        assert alias_test_false(false)
        assert not alias_test_true(false)
    assert alias_test_true("on")
    assert alias_test_true("yes")
    assert alias_test_true(1)
    assert alias_test_true("anything-not-falsey")


def test_duck_object():
    ducklike = DuckObject(walk="waddle", talk="quack")
    assert ducklike.walk == "waddle"
    assert ducklike.talk == "quack"
    assert not hasattr(ducklike, "bill")


@pytest.mark.skip(
    reason="no way of currently testing this without actual Unicat connection"
)
def test_field_columns():
    assert True


def test_convert_value_to_fielddata():
    assert convert_value_to_fielddata("text", "123") == "123"
    assert convert_value_to_fielddata("textline", "123") == "123"
    assert convert_value_to_fielddata("textline", 123) == "123"
    assert convert_value_to_fielddata("textlist", "123\n456") == ["123", "456"]
    assert convert_value_to_fielddata("number", "123") == 123
    assert convert_value_to_fielddata("decimal", "123") == 123
    assert convert_value_to_fielddata("decimal", "123.45") == 123.45
    assert convert_value_to_fielddata("decimal", "123,45") == 123.45
    assert convert_value_to_fielddata("boolean", "123") == True
    assert convert_value_to_fielddata("code", "123") == "123"
    assert convert_value_to_fielddata("barcode", "123") == "123"
    assert convert_value_to_fielddata("class", '{"abc": "123"}') == {"abc": "123"}
    assert convert_value_to_fielddata(
        "classlist", '[{"abc": "123"},{"def": "456"}]'
    ) == [{"abc": "123"}, {"def": "456"}]
    assert convert_value_to_fielddata("image", "<gid-123>") == "<gid-123>"
    assert convert_value_to_fielddata("imagelist", "<gid-123>\n<gid-456>") == [
        "<gid-123>",
        "<gid-456>",
    ]
    assert convert_value_to_fielddata("file", "<gid-123>") == "<gid-123>"
    assert convert_value_to_fielddata("filelist", "<gid-123>\n<gid-456>") == [
        "<gid-123>",
        "<gid-456>",
    ]
    assert convert_value_to_fielddata("record", "<gid-123>") == "<gid-123>"
    assert convert_value_to_fielddata("recordlist", "<gid-123>\n<gid-456>") == [
        "<gid-123>",
        "<gid-456>",
    ]
    assert convert_value_to_fielddata("fieldpicker", "<gid-123>") == "<gid-123>"


def test_convert_value_to_fielddata_errors():
    with pytest.raises(Exception) as e:
        convert_value_to_fielddata("class", "abc")
    with pytest.raises(Exception) as e:
        convert_value_to_fielddata("classlist", "abc")


def test_convert_fielddata_to_value():
    assert convert_fielddata_to_value("text", "123") == "123"
    assert convert_fielddata_to_value("textline", "123") == "123"
    assert convert_fielddata_to_value("textlist", ["123", "456"]) == "123\n456"
    assert convert_fielddata_to_value("number", 123) == 123
    assert convert_fielddata_to_value("decimal", 123) == 123
    assert convert_fielddata_to_value("decimal", 123.45) == 123.45
    assert convert_fielddata_to_value("boolean", True) == "yes"
    assert convert_fielddata_to_value("code", "123") == "123"
    assert convert_fielddata_to_value("barcode", "123") == "123"
    assert convert_fielddata_to_value("class", {"abc": "123"}) == '{\n  "abc": "123"\n}'
    assert (
        convert_fielddata_to_value("classlist", [{"abc": "123"}, {"def": "456"}])
        == '[\n  {\n    "abc": "123"\n  },  {\n    "def": "456"\n  }\n]'
    )
    assert convert_fielddata_to_value("image", "<gid-123>") == "<gid-123>"
    assert (
        convert_fielddata_to_value("imagelist", ["<gid-123>", "<gid-456>"])
        == "<gid-123>\n<gid-456>"
    )
    assert convert_fielddata_to_value("file", "<gid-123>") == "<gid-123>"
    assert (
        convert_fielddata_to_value("filelist", ["<gid-123>", "<gid-456>"])
        == "<gid-123>\n<gid-456>"
    )
    assert convert_fielddata_to_value("record", "<gid-123>") == "<gid-123>"
    assert (
        convert_fielddata_to_value("recordlist", ["<gid-123>", "<gid-456>"])
        == "<gid-123>\n<gid-456>"
    )
    assert convert_fielddata_to_value("fieldpicker", "<gid-123>") == "<gid-123>"


def test_diff_record_fields_data(unicat):
    record = unicat.get_root_record()
    localizedfielddata = {
        "en": {"field-1-name": "Field 1 value 0"},
        "nl": {"field-1-name": "Veld 1 waarde 0"},
    }
    diff = diff_record_fields_data(unicat, record, localizedfielddata)
    assert not len(diff)
    localizedfielddata = {
        "en": {"field-1-name": "Field 1 value CHANGED"},
        "nl": {"field-1-name": "Veld 1 waarde 0"},
    }
    diff = diff_record_fields_data(unicat, record, localizedfielddata)
    assert len(diff) == 1
    assert "en" in diff
    assert "nl" not in diff
    localizedfielddata = {
        "en": {"field-1-name": "Field 1 value CHANGED"},
        "nl": {"field-1-name": "Veld 1 waarde AANGEPAST"},
    }
    diff = diff_record_fields_data(unicat, record, localizedfielddata)
    assert len(diff) == 2
    assert "en" in diff
    assert "nl" in diff


def test_merge_dicts():
    a = {"a": 1, "c": 99}
    x = merge_dicts(a, {"b": 2, "c": 3})
    assert x["a"] == a["a"] == 1
    assert x["b"] == a["b"] == 2
    assert x["c"] == a["c"] == 3
    a = {"a": 1, "c": 99}
    x = merge_dicts(dict(a), {"b": 2, "c": 3})
    assert x["a"] == a["a"] == 1
    assert x["b"] == 2
    assert x["c"] == 3
    assert a["c"] == 99


def check_hash(hash):
    if len(hash) != 8:
        return False
    for char in hash:
        if char not in "0123456789abcdef":
            return False
    return True


def test_check_hash():
    assert check_hash("1234abcd")
    assert not check_hash("1234ABCD")
    assert not check_hash("1234GHIJ")
    assert not check_hash("1234abc")
    assert not check_hash("1234abcde")


def test_hash_text():
    assert check_hash(hash_text("abcdef"))
    assert check_hash(hash_text("🧐🥸"))


def test_hash_data():
    assert check_hash(hash_data([1, 2, 3]))
    assert check_hash(hash_data({"a": "b", "c": [1, 2, 3]}))



from pathlib import Path
from unicat import Unicat

connection = {
    "server": "http://cc.localhost",
    "project_gid": "84a1b5e4-fb5d-4ad6-a497-a3cf5eaf29b1",
    "api_key": "pynNf5w^pyF9PephC^d574N#pW3NN7qF",
}
CMS_GID = "99388f36-1707-4e22-97c3-874fdc8c1869"

folder = Path("files")
if not Path(folder).is_absolute():
    folder = Path(Path.cwd(), folder)
folder = folder.resolve()

unicat = Unicat(
    connection["server"],
    connection["project_gid"],
    connection["api_key"],
    folder,
)
if not unicat.connect():
    raise Exception("Invalid connection settings")

project = unicat.project
print(project.name)
# print(project.owner.username)
# print(project.default_language, project.languages)
# print(project.channels)
# print(project.orderings)
# print(project.fieldlists)
print()

channels = [project.channels["Website NL"]]

record = unicat.get_record(CMS_GID)
result = unicat.mutate.copy_record_channels_down(record, channels)
print(result)
print()

job = unicat.mutate.copy_record_channels_down(record, channels, return_job=True)
print(job, job.progress, job.return_value)
print()

job = unicat.mutate.copy_record_channels_down(record, channels, return_job=True)
for status in job.track():
    assert status == job.status
    print(job.name, job.status)
print(job, job.progress, job.return_value)
print()



class UnicatProjectBackup:
    def __init__(self, unicat, project_gid, backup):
        self._unicat = unicat
        self._project = self._unicat.projects[project_gid]
        self._data = backup

    @property
    def version(self):
        return self._data["version"]

    @property
    def name(self):
        return self._data["name"]

    @property
    def created_by(self):
        return self._data["created_by"]

    @property
    def timestamp(self):
        return self._data["timestamp"]



# from .project import UnicatProject


class UnicatUser:
    def __init__(self, unicat, gid):
        self._unicat = unicat
        self._data = self._unicat.users[gid]

    @property
    def gid(self):
        return self._data["gid"]

    @property
    def username(self):
        return self._data["username"]

    @property
    def name(self):
        return self._data["name"]

    @property
    def avatar(self):
        return self._data["avatar"]



class UnicatAsset:
    def __init__(self, unicat, gid):
        self._unicat = unicat
        self._data = self._unicat.assets[gid]

    @property
    def gid(self):
        return self._data["gid"]

    @property
    def pathname(self):
        return self._data["pathname"]

    @property
    def path(self):
        return self._data["path"]

    @property
    def name(self):
        return self._data["name"]

    @property
    def is_file(self):
        return self._data["is_file"]

    @property
    def type(self):
        return self._data["type"]

    @property
    def childcount(self):
        return self._data["childcount"]

    @property
    def status(self):
        return self._data["status"]

    @property
    def is_deleted(self):
        return self._data["status"] == "deleted"

    @property
    def info(self):
        return self._data["info"]

    @property
    def transforms(self):
        return self._data["transforms"]

    @property
    def default_transform(self):
        if self._data["transforms"] is None:
            return None
        return self._data["transforms"].get("_main_", None)

    @property
    def title(self):
        return self._data["title"]

    @property
    def description(self):
        return self._data["description"]

    @property
    def updated_on(self):
        return self._data["updated_on"]

    def publish(self):
        if not self.is_file:
            return None
        public_url = self._unicat.publish_asset(self)
        return public_url

    def publish_transformed(self, transform=None):
        if not self.is_file:
            return None
        public_url = self._unicat.publish_transformed_asset(self, transform=transform)
        return public_url

    def download(self, pathname=None):
        if not self.is_file:
            return None
        pathname = self._unicat.download_asset(
            self, pathname=pathname, updated_on=self.updated_on
        )
        return pathname

    def download_transformed(self, transform=None, pathname=None):
        if not self.is_file:
            return None
        pathname = self._unicat.download_transformed_asset(
            self, transform=transform, pathname=pathname, updated_on=self.updated_on
        )
        return pathname



from unicat.utils import hash_data
from datetime import datetime

# Data as it is returned from Unicat, then stored in the UnicatApi
MockProjectData = {
    "cc.users": {
        "<user-1>": {
            "gid": "<user-1>",
            "username": "user-1",
            "name": "User the First",
            "avatar": "user-1-avatar.png",
        }
    },
    "cc.projects": {
        "<project-1>": {
            "gid": "<project-1>",
            "name": "Project the First",
            "icon": "project-1-icon.jpg",
            "owner": "<user-1>",
            "status": "active",
            "options": {
                "timezone": "Europe/Amsterdam",
                "languages": [
                    "en",
                    "nl",
                ],
                "channels": {
                    "__all__": "[all records]",
                    "<channel-1>": "Channel the First",
                    "<channel-2>": "Channel the Second",
                },
                "orderings": {
                    "<ordering-1>": "Ordering the First",
                },
                "fieldlists": {
                    "<fieldlist-1>": "Fieldlist the First",
                },
                "orderedchannels": [
                    "__all__",
                    "<channel-1>",
                    "<channel-2>",
                ],
                "orderedorderings": [
                    "<ordering-1>",
                ],
                "orderedfieldlists": [
                    "<fieldlist-1>",
                ],
            },
            "backups": {
                "versions": [
                    {
                        "version": 3,
                        "name": "Backup 3",
                        "created_by": "Unicat",
                        "timestamp": 1610635129.04762,
                    },
                    {
                        "version": 2,
                        "name": "Backup 2",
                        "created_by": "Unicat",
                        "timestamp": 1610635128.04762,
                    },
                    {
                        "version": 1,
                        "name": "Backup 1",
                        "created_by": "Unicat",
                        "timestamp": 1610635127.04762,
                    },
                    {
                        "version": 0,
                        "name": "Initial snapshot",
                        "created_by": "Unicat",
                        "timestamp": 1610635126.04762,
                    },
                ],
                "next_version": 4,
            },
            "build": 1,
        },
    },
    "cc.projects_members": {
        "<project-1>/<user-1>": {
            "project_gid": "<project-1>",
            "user_gid": "<user-1>",
            "status": "active",
            "roles": ["owner"],
            "options": {},
        }
    },
    "cc.languages": {"en": "English", "nl": "Dutch", "it": "Italian"},
    "records": {
        "<record-0>": {
            "gid": "<record-0>",
            "parent": None,
            "canonical": "<record-0>",
            "backlinks": [],
            "treelevel": 1,
            "path": ["<record-0>"],
            "childcount": 1,
            "definition": "<definition-1>",
            "title": {"nl": "Record de Root", "en": "Record the Root"},
            "status": "published",
            "channels": [
                "__all__",
                "<channel-1>",
            ],
            "orderings": {
                "<ordering-1>": 1,
            },
            "is_link": False,
            "fields": {
                "nl": {"field-1-name": "Veld 1 waarde 0"},
                "en": {"field-1-name": "Field 1 value 0"},
            },
        },
        "<record-1>": {
            "gid": "<record-1>",
            "parent": "<record-0>",
            "canonical": "<record-1>",
            "backlinks": [],
            "treelevel": 2,
            "path": ["<record-1>", "<record-0>"],
            "childcount": 0,
            "definition": "<definition-1>",
            "title": {"nl": "Record de Eerste", "en": "Record the First"},
            "status": "published",
            "channels": [
                "__all__",
                "<channel-1>",
            ],
            "orderings": {
                "<ordering-1>": 1,
            },
            "is_link": False,
            "fields": {
                "nl": {"field-1-name": "Veld 1 waarde 1"},
                "en": {"field-1-name": "Field 1 value 1"},
            },
        },
    },
    "definitions": {
        "<definition-1>": {
            "gid": "<definition-1>",
            "original": None,
            "name": "definition-1-name",
            "label": {"nl": "Definitie 1", "en": "Definition 1"},
            "classes": ["<class-1>"],
            "fields": ["<field-1>"],
            "titlefield": "<field-1>",
            "fieldlists": {"<fieldlist-1>": ["<field-1>"]},
            "layout": "<layout-1>",
            "childdefinitions": [],
            "is_base": False,
            "is_new": False,
            "is_working_copy": False,
            "is_extended": False,
        }
    },
    "classes": {
        "<class-1>": {
            "gid": "<class-1>",
            "name": "class-1-name",
            "original": None,
            "label": {"nl": "Klasse 1", "en": "Class 1"},
            "fields": ["<field-1>"],
            "layout": "<layout-2>",
            "is_new": False,
            "is_working_copy": False,
        }
    },
    "fields": {
        "<field-1>": {
            "gid": "<field-1>",
            "name": "field-1-name",
            "original": None,
            "label": {"nl": "Veld 1", "en": "Field 1"},
            "type": "text",
            "options": {},
            "is_localized": True,
            "is_required": False,
            "initial": {"nl": "", "en": ""},
            "unit": "m²",
            "is_new": False,
            "is_working_copy": False,
        },
        "<field-2>": {
            "gid": "<field-2>",
            "name": "field-2-name",
            "original": None,
            "label": {"nl": "Veld 2", "en": "Field 2"},
            "type": "text",
            "options": {},
            "is_localized": True,
            "is_required": False,
            "initial": {"nl": "", "en": ""},
            "unit": None,
            "is_new": False,
            "is_working_copy": False,
        },
    },
    "layouts": {
        "<layout-1>": {
            "gid": "<layout-1>",
            "name": "layout-1-name",
            "original": None,
            "root": "<component-1>",
            "components": {"<component-1>": {"type": "vertical", "content": []}},
            "gid_map": {},
            "is_new": False,
            "is_working_copy": False,
        },
        "<layout-2>": {
            "gid": "<layout-2>",
            "name": "layout-2-name",
            "original": None,
            "root": "<component-2>",
            "components": {"<component-2>": {"type": "vertical", "content": []}},
            "gid_map": {},
            "is_new": False,
            "is_working_copy": False,
        },
    },
    "assets": {
        "<asset-0>": {
            "gid": "<asset-0>",
            "pathname": "/",
            "is_file": False,
            "type": None,
            "childcount": 1,
            "status": "published",
            "info": {"filecount": 1, "foldercount": 0},
            "transforms": None,
            "path": "/",
            "name": "",
            "title": {"nl": "Asset de Root", "en": "Asset the Root"},
            "description": {"nl": "", "en": ""},
            "updated_on": 1610635126.04762,
        },
        "<asset-1>": {
            "gid": "<asset-1>",
            "pathname": "/asset-1-name.svg",
            "is_file": True,
            "type": "svg",
            "childcount": 0,
            "status": "published",
            "info": {
                "pages": 1,
                "width": 333,
                "colors": "truecolormatte",
                "format": "svg",
                "height": 333,
                "animated": False,
                "filesize": 1192,
                "metadata": {
                    "date:create": "2021-01-14T15:38:43+01:00",
                    "date:modify": "2021-01-14T15:38:43+01:00",
                },
                "colorspace": "srgb",
                "resolution": [300, 300],
                "transparent": None,
            },
            "transforms": {
                "_main_": {
                    "crop": None,
                    "hotspot": [0.6467065868263473, 0.40119760479041916],
                }
            },
            "path": "/",
            "name": "asset-1-name.svg",
            "title": {
                "nl": "Een generieke afbeelding",
                "en": "A generic image",
            },
            "description": {
                "nl": "Afbeelding - een lijntekening van een lijst met daarin bergen en de zon.",
                "en": "Image - a line drawing of a frame containing mountains and the sun.",
            },
            "updated_on": 1610635123.351925,
        },
    },
    "queries": {
        "<query-1>": {
            "gid": "<query-1>",
            "type": "record",
            "name": "query-1-record-name",
            "q": "",
            "filter": ["and", "", [["validation", "not_translated"]]],
        },
        "<query-2>": {
            "gid": "<query-2>",
            "type": "asset",
            "name": "query-2-asset-name",
            "q": "",
            "filter": ["and", "", [["validation", "any"]]],
        },
        "<query-3>": {
            "gid": "<query-3>",
            "type": "schema",
            "name": "query-3-schema-name",
            "q": "",
            "filter": ["and", "", [["is_committed", "is_true"]]],
        },
    },
}


class MockApi:
    """Used to simulate regular operation of the API."""

    def __init__(self, base_url, project_gid, asset_folder=None):
        self._base_url = base_url
        self._project_gid = project_gid
        self._asset_folder = asset_folder
        self.data = MockProjectData

    def connect(self, secret_api_key):
        return self._call("/auth")

    def init(self):
        return self._call("/init")

    def call(self, endpoint, data=None, method="post"):
        return self._call(endpoint, data=data)

    def upload(self, endpoint, data=None, localfilepath=None):
        return self._call_upload(endpoint, data=data)

    def publish(self, asset):
        return self._call_dam(asset)

    def transform(self, asset, options=None):
        return self._call_dam(asset)

    def download(self, asset, pathname=None, updated_on=None):
        return self._call_media(asset)

    def download_transformed(self, asset, options=None, pathname=None, updated_on=None):
        return self._call_media(asset)

    def sync(self):
        success, result = self._call_sync()
        if not success:
            return None

    def prepare_for_job(self):
        success, result = self._call_sync()
        if not success:
            return None
        return 1  # cc-cursor

    def track_job(
        self, cursor, job_gid, timeout_in_seconds=None, poll_interval_in_seconds=1.0
    ):
        now = datetime.now().timestamp()
        progress = {
            "gid": job_gid,
            "name": self.name,
            "status": "queued",
            "info": {},
            "created_on": now,
            "updated_on": now,
        }
        yield progress
        progress["status"] = "processing"
        progress["updated_on"] = now + 1
        yield progress
        progress["status"] = "processing"
        progress["updated_on"] = now + 1
        yield progress
        progress["status"] = "done"
        progress["updated_on"] = now + 1
        yield progress

    def project_member_key(self, project_member):
        return f"{project_member['project_gid']}/{project_member['user_gid']}"

    def _json_response(self, result=None):
        return True, result

    def _response_error(self, code=418, message="Mock failure", info=None):
        return False, {"code": int(code), "message": message, "info": info}

    def _call(self, endpoint, data=None):
        hash = hash_data([endpoint, data])
        results = {
            hash_data(["/records/root", None]): {"root": "<record-0>"},
            hash_data(["/records/get", {"records": ["<record-999>"]}]): {"records": []},
            hash_data(["/assets/root", None]): {"root": "<asset-0>"},
            hash_data(["/assets/get", {"assets": ["<asset-999>"]}]): {"assets": []},
            hash_data(["/assets/get", {"pathname": "asset-1-pathname"}]): {
                "asset": "<asset-1>"
            },
        }
        return self._json_response(results[hash] if hash in results else None)

    def _call_sync(self):
        return self._json_response()

    def _call_upload(self, endpoint, data=None):
        return self._json_response()

    def _call_dam(self, asset):
        return self._json_response(
            {"public_url": "mocks://unicat.app/p/src/any-filename.ext"}
        )

    def _call_media(self, asset):
        return "/tmp/unicat/any-filename.ext"


class MockApiFailure(MockApi):
    """Used to simulate (expected or unexpected) errors when accessing the API."""

    def _call(self, endpoint, data=None):
        return self._response_error()

    def _call_sync(self):
        return self._response_error()

    def _call_upload(self, endpoint, data=None):
        return self._response_error()

    def _call_dam(self, asset):
        return self._response_error()

    def _call_media(self, asset):
        return False



class UnicatField:
    def __init__(self, unicat, gid):
        self._unicat = unicat
        self._data = self._unicat.fields[gid]

    @property
    def gid(self):
        return self._data["gid"]

    @property
    def original(self):
        if not self._data["original"]:
            return None
        return UnicatField(self._unicat, self._data["original"])

    @property
    def name(self):
        return self._data["name"]

    @property
    def type(self):
        return self._data["type"]

    @property
    def class_(self):
        if self._data["type"] not in ("class", "classlist"):
            return None

        from .class_ import UnicatClass

        return UnicatClass(self._unicat, self._data["options"]["class"])

    @property
    def options(self):
        return self._data["options"]

    @property
    def is_localized(self):
        return self._data["is_localized"]

    @property
    def is_required(self):
        return self._data["is_required"]

    @property
    def label(self):
        return self._data["label"]

    @property
    def initial(self):
        return self._data["initial"]

    @property
    def unit(self):
        return self._data["unit"]

    @property
    def title(self):
        if not self.unit:
            return self.label
        return {
            language: f"{label} [{self.unit}]" for language, label in self.label.items()
        }

    @property
    def is_new(self):
        return self._data["is_new"]

    @property
    def is_working_copy(self):
        return self._data["is_working_copy"]

    @property
    def is_committed(self):
        return not self._data["is_new"] and not self._data["is_working_copy"]



class UnicatLayout:
    def __init__(self, unicat, gid):
        self._unicat = unicat
        self._data = self._unicat.layouts[gid]

    @property
    def gid(self):
        return self._data["gid"]

    @property
    def original(self):
        if not self._data["original"]:
            return None
        return UnicatLayout(self._unicat, self._data["original"])

    @property
    def name(self):
        return self._data["name"]

    @property
    def root(self):
        return self._data["root"]

    @property
    def components(self):
        return self._data["components"]

    @property
    def is_new(self):
        return self._data["is_new"]

    @property
    def is_working_copy(self):
        return self._data["is_working_copy"]



from collections import defaultdict
from .record_field import UnicatRecordField


class UnicatRecord:
    def __init__(self, unicat, gid, validation_report=None):
        self._unicat = unicat
        self._data = self._unicat.records[gid]
        fields_by_name = {}
        for field in self.definition.all_fields:
            fields_by_name[field.name] = field
        self._fields = {
            language: {
                fieldname: UnicatRecordField(
                    self._unicat,
                    self,
                    fields_by_name[fieldname],
                    self._data["fields"][language][fieldname],
                )
                for fieldname in self._data["fields"][language]
            }
            for language in self._data["fields"]
        }
        self.validation_report = validation_report

    @property
    def gid(self):
        return self._data["gid"]

    @property
    def canonical(self):
        return self._data["canonical"]

    @property
    def parent(self):
        return self._data["parent"]

    @property
    def backlinks(self):
        return self._data["backlinks"]

    @property
    def is_link(self):
        return self._data["is_link"]

    @property
    def is_deleted(self):
        return self._data["status"] == "deleted"

    @property
    def treelevel(self):
        return self._data["treelevel"]

    @property
    def path(self):
        return self._data["path"]

    @property
    def title(self):
        return self._data["title"]

    @property
    def channels(self):
        return self._data["channels"]

    @property
    def orderings(self):
        return self._data["orderings"]

    @property
    def childcount(self):
        return self._data["childcount"]

    @property
    def definition(self):
        from .definition import UnicatDefinition

        return UnicatDefinition(self._unicat, self._data["definition"])

    @property
    def fields(self):
        return self._fields



class UnicatError(Exception):
    def __init__(self, message: str, error_result: dict):
        if type(error_result) != dict:
            raise TypeError("error_result must be a Unicat error result dict")
        super().__init__(message, error_result)

    @property
    def code(self):
        return self.args[1].get("code", None)

    @property
    def message(self):
        return self.args[1].get("message", None)

    @property
    def info(self):
        return self.args[1].get("info", None)

    def __str__(self):
        return self.args[0]



class UnicatRecordField:
    """A record field can have a value, a reference (record, asset), or it can be
    nested for class-fields.
    We also support 'list' versions of these.

    # for values

    artnr = record.fields[language]["artnr"]  # a recordfield
    artnr.value             # "CMS 225-945"
    artnr.field.label       # "Article number"
    artnr.field.type        # "textline"

    for references

    image = record.fields[language]["image"]  # both recordfield and asset
    image.value             # "a0a80c9c-fa1b-4573-ac98-b7b07c81b583"
    image.field.label       # "Image"
    image.pathname          # "/products/cms225.eps"

    for class fields

    dimensions_interior = record.fields[language]["dimensions_interior"]
                            # a recordfield and classfield
    dimensions_interior.value               # {"width__mm": 374, …}
    dimensions_interior["width__mm"].value  # 374  -- this is a recordfield

    # for list values

    colors = record.fields[language]["colors"]
    colors.value            # ["Red", "Blue"]
    colors.field.label      # "Colors"
    colors[0].value         # "Red" -- this is just a string

    # for list references

    images = record.fields[language]["images"]
    images.value            # ["a0a80c9c-fa1b-4573-ac98-b7b07c81b583", ]
    images.field.label      # "Images"
    images[0]               # this is just an asset
    images[0].pathname      # "/products/cms225.eps"

    for classlist fields

    tablespecs = record.fields[language]["tablespecs"]
    tablespecs.value                        # [{"width__mm": 7, …}, …]
    tablespecs.field.label                  # "Table specs"
    tablespecs[0]                           # this is recordfield-like
    tablespecs[0]["width__mm"].value        # 7
    tablespecs[0]["width__mm"].field.label  # "Width"

    ## TODO:

    change the fluent interface around a bit, so a recordfield is more like a field,
    with field.name, field.label, and it still stores the value in field.value -- but,
    for references, we need to go through the value, so you would get field.value.pathname
    for an image reference, or field.value[0].pathname for an imagelist item.

    old                         value                   new
    ---                         -----                   ---
    recordfield.field.name      images                  recordfield.name
    recordfield.field.type      imagelist               recordfield.type
    recordfield.value           [<gid>, <gid, ...]      recordfield.value (kinda)
    recordfield[0].pathname     /products/image.png     recordfield.value[0].pathname

    if you use it like

    image = record.fields[language]["image"]
    print(image.field.label, image.pathname, image.value)


    then it looks kinda ok, but if you loop through the fields it gets a bit weird

    for field in fields:
        print(field.field.label, field.pathname, field.value)

    the new way would look like this:

    image = record.fields[language]["image"]
    print(image.label, image.value.pathname, image.value.gid)

    for field in fields:
        print(field.label, field.value.pathname, field.value.gid)

    it's more clear that the value is of type image, and the field is more field-like.

    for imagelists:

    images = record.fields[language]["image"]
    print(images.field.label, images[0].pathname, images[0])

    becomes

    images = record.fields[language]["image"]
    print(images.label, images.value[0].pathname, images.value[0].gid)
    """

    def __init__(self, unicat, record, field, value):
        self._unicat = unicat
        self._record = record
        self.field = field
        self.value = value

    def __str__(self):
        return str(self.value)

    def __bool__(self):
        return self.value is not None

    def __getattr__(self, name):
        # used for fluent interface for references
        # lazily loads the record or asset
        if self.field.type == "record":
            ref = self._unicat.get_record(self.value)
            return getattr(ref, name)
        elif self.field.type in ["file", "image"]:
            ref = self._unicat.get_asset(self.value)
            return getattr(ref, name)
        raise AttributeError(name)

    def __getitem__(self, key):
        # used for lists (int key) and class fields (str key)
        if isinstance(key, int):
            if self.value is None:
                raise StopIteration  # so enumerate() keeps working on None too
            if self.field.type == "recordlist":
                refkey = self.value[key]
                return self._unicat.get_record(refkey)
            elif self.field.type in ("filelist", "imagelist"):
                refkey = self.value[key]
                return self._unicat.get_asset(refkey)
            elif self.field.type == "classlist":
                classdata = self.value[key]
                return self._get_classfields(classdata)
            return self.value[key]
        elif isinstance(key, str):
            classfields = self._get_classfields(self.value)
            return classfields[key]
        raise KeyError(key)

    def _get_classfields(self, data):
        classfields = {}
        if not data or not self.field.class_:
            return classfields
        from .record_class_field import UnicatRecordClassField

        for field in self.field.class_.fields:
            classfields[field.name] = UnicatRecordClassField(
                self._unicat,
                self,
                field,
                data[field.name] if field.name in data else None,
            )
        return classfields



# from .project import UnicatProject


class UnicatJob:
    def __init__(self, unicat, cursor, gid, return_value):
        self._unicat = unicat
        self._cursor = cursor
        self._gid = gid
        self._progress = {}
        self._return_value = return_value

    @property
    def gid(self):
        return self._gid

    @property
    def name(self):
        if "job" not in self._progress:
            return "unknown"
        return self._progress["job"]

    @property
    def status(self):
        if "status" not in self._progress:
            return "untracked"
        return self._progress["status"]

    @property
    def info(self):
        if "info" not in self._progress:
            return {}
        return self._progress["info"]

    @property
    def created_on(self):
        if "created_on" not in self._progress:
            return None
        return self._progress["created_on"]

    @property
    def updated_on(self):
        if "updated_on" not in self._progress:
            return None
        return self._progress["updated_on"]

    @property
    def progress(self):
        return {
            "gid": self.gid,
            "name": self.name,
            "status": self.status,
            "info": self.info,
            "created_on": self.created_on,
            "updated_on": self.updated_on,
        }

    @property
    def return_value(self):
        return self._return_value

    def track(self, timeout_in_seconds=None, poll_interval_in_seconds=1.0):
        for progress in self._unicat.api.track_job(
            self._cursor,
            self._gid,
            timeout_in_seconds=timeout_in_seconds,
            poll_interval_in_seconds=poll_interval_in_seconds,
        ):
            self._progress = progress
            yield self.status



import pytest
from unicat import UnicatError
from unicat.utils import DuckObject
from setup import unicat, unicaterror

# we test most methods with unicaterror, but that will test the code-paths for
# arguments &c.


def test_mutate(unicat):
    assert hasattr(unicat, "mutate")


def test_mutate_project_settings(unicaterror):
    with pytest.raises(UnicatError):
        unicaterror.mutate.add_language("it")
    with pytest.raises(UnicatError):
        unicaterror.mutate.remove_language("it")
    with pytest.raises(UnicatError):
        unicaterror.mutate.create_channel("New Channel Name")
    with pytest.raises(UnicatError):
        unicaterror.mutate.delete_channel(unicaterror.project.default_channel)
    with pytest.raises(UnicatError):
        unicaterror.mutate.create_ordering("New Ordering Name")
    with pytest.raises(UnicatError):
        unicaterror.mutate.delete_ordering(unicaterror.project.default_ordering)
    with pytest.raises(UnicatError):
        unicaterror.mutate.create_fieldlist("New Fieldlist Name")
    with pytest.raises(UnicatError):
        unicaterror.mutate.delete_fieldlist(unicaterror.project.default_fieldlist)


def test_mutate_definitions(unicaterror):
    language = unicaterror.project.default_language
    fieldlist = unicaterror.project.default_fieldlist
    definition = unicaterror.get_definition("<definition-1>")
    duckdefinition = DuckObject(gid="<definition-duck>")
    class_ = unicaterror.get_class("<class-1>")
    field = unicaterror.get_field("<field-1>")
    with pytest.raises(UnicatError):
        unicaterror.mutate.create_definition(
            name="new",
            label={language: "New"},
            classes=[class_],
            fields=[field],
            titlefield=field,
            childdefinitions=[],
        )
    with pytest.raises(UnicatError):
        unicaterror.mutate.create_definition(
            other="layout - know what you're doing!",
        )
    with pytest.raises(UnicatError):
        unicaterror.mutate.modify_definition(
            definition,
            name="new",
            label={language: "New"},
            classes=[class_],
            fields=[field],
            titlefield=field,
            childdefinitions=[],
        )
    with pytest.raises(UnicatError):
        unicaterror.mutate.modify_definition(
            definition,
            other="layout - know what you're doing!",
        )
    with pytest.raises(UnicatError):
        # also - know what you're doing!
        unicaterror.mutate.modify_definition_modify_layout(
            definition,
            name="layout",
            root="<gid>",
            components={"<gid>": {"type": "vertical", "content": []}},
        )
    with pytest.raises(UnicatError):
        unicaterror.mutate.modify_definition_add_class(definition, class_)
    with pytest.raises(UnicatError):
        unicaterror.mutate.modify_definition_remove_class(definition, class_)
    with pytest.raises(UnicatError):
        unicaterror.mutate.modify_definition_add_field(definition, field)
    with pytest.raises(UnicatError):
        unicaterror.mutate.modify_definition_remove_field(definition, field)
    with pytest.raises(UnicatError):
        unicaterror.mutate.modify_definition_fieldlist_add_field(
            definition, fieldlist, field
        )
    with pytest.raises(UnicatError):
        unicaterror.mutate.modify_definition_fieldlist_remove_field(
            definition, fieldlist, field
        )
    with pytest.raises(UnicatError):
        unicaterror.mutate.modify_definition_add_childdefinition(
            definition, childdefinition=duckdefinition
        )
    with pytest.raises(UnicatError):
        unicaterror.mutate.modify_definition_remove_childdefinition(
            definition, childdefinition=duckdefinition
        )
    with pytest.raises(UnicatError):
        unicaterror.mutate.commit_definition(duckdefinition)
    with pytest.raises(UnicatError):
        unicaterror.mutate.save_as_new_definition(duckdefinition)
    with pytest.raises(UnicatError):
        unicaterror.mutate.delete_definition(definition)


def test_mutate_classes(unicaterror):
    language = unicaterror.project.default_language
    class_ = unicaterror.get_class("<class-1>")
    duckclass = DuckObject(gid="<class-duck>")
    field = unicaterror.get_field("<field-1>")
    with pytest.raises(UnicatError):
        unicaterror.mutate.create_class(
            name="new", label={language: "New"}, fields=[field]
        )
    with pytest.raises(UnicatError):
        unicaterror.mutate.create_class(
            other="layout - know what you're doing!",
        )
    with pytest.raises(UnicatError):
        unicaterror.mutate.modify_class(
            class_, name="new", label={language: "New"}, fields=[field]
        )
    with pytest.raises(UnicatError):
        unicaterror.mutate.modify_class(
            class_,
            other="layout - know what you're doing!",
        )
    with pytest.raises(UnicatError):
        # also - know what you're doing!
        unicaterror.mutate.modify_class_modify_layout(
            class_,
            name="layout",
            root="<gid>",
            components={"<gid>": {"type": "vertical", "content": []}},
        )
    with pytest.raises(UnicatError):
        unicaterror.mutate.modify_class_add_field(class_, field)
    with pytest.raises(UnicatError):
        unicaterror.mutate.modify_class_remove_field(class_, field)
    with pytest.raises(UnicatError):
        unicaterror.mutate.commit_class(duckclass)
    with pytest.raises(UnicatError):
        unicaterror.mutate.save_as_new_class(duckclass)
    with pytest.raises(UnicatError):
        unicaterror.mutate.delete_class(class_)


def test_mutate_fields(unicaterror):
    language = unicaterror.project.default_language
    field = unicaterror.get_field("<field-1>")
    duckfield = DuckObject(gid="<field-duck>")
    with pytest.raises(UnicatError):
        unicaterror.mutate.create_field(
            name="new",
            type="text",
            is_localized=True,
            is_required=False,
            label={language: "New"},
            unit="mm²",
            initial={},
            options={},
        )
    with pytest.raises(UnicatError):
        unicaterror.mutate.modify_field(
            field,
            name="new",
            type="text",
            is_localized=True,
            is_required=False,
            label={language: "New"},
            unit="mm²",
            initial={},
            options={},
        )
    with pytest.raises(UnicatError):
        unicaterror.mutate.commit_field(duckfield)
    with pytest.raises(UnicatError):
        unicaterror.mutate.save_as_new_field(duckfield)
    with pytest.raises(UnicatError):
        unicaterror.mutate.delete_field(field)


def test_mutate_records(unicaterror):
    language = unicaterror.project.default_language
    channel = unicaterror.project.default_channel
    ordering = unicaterror.project.default_ordering
    fieldlist = unicaterror.project.default_fieldlist
    record = unicaterror.get_record("<record-1>")
    parentrecord = unicaterror.get_record("<record-0>")
    definition = unicaterror.get_definition("<definition-1>")
    class_ = unicaterror.get_class("<class-1>")
    field = unicaterror.get_field("<field-1>")
    with pytest.raises(UnicatError):
        unicaterror.mutate.create_record(parentrecord, ordering=ordering)
    with pytest.raises(UnicatError):
        unicaterror.mutate.set_record_definition(record, definition)
    with pytest.raises(UnicatError):
        unicaterror.mutate.extend_record_definition_add_class(record, class_)
    with pytest.raises(UnicatError):
        unicaterror.mutate.extend_record_definition_add_field(record, field)
    with pytest.raises(UnicatError):
        unicaterror.mutate.extend_record_definition_add_fieldlist_field(
            record, fieldlist, field
        )
    with pytest.raises(UnicatError):
        unicaterror.mutate.update_record(
            record, localizedfielddata={language: {"field-1-name": "value"}}
        )
    with pytest.raises(UnicatError):
        unicaterror.mutate.set_record_channels(record, channels=[channel], enabled=True)
    with pytest.raises(UnicatError):
        unicaterror.mutate.copy_record_channels_from_parent(record, channels=[channel])
    with pytest.raises(UnicatError):
        unicaterror.mutate.copy_record_channels_down(record, channels=[channel])
    with pytest.raises(UnicatError):
        unicaterror.mutate.copy_record_channels_up(record, channels=[channel])
    with pytest.raises(UnicatError):
        unicaterror.mutate.set_record_orderings(record, orderings={ordering: 123})
    with pytest.raises(UnicatError):
        unicaterror.mutate.link_record(parentrecord, record, ordering=ordering)
    with pytest.raises(UnicatError):
        unicaterror.mutate.delete_record(record)
    with pytest.raises(UnicatError):
        unicaterror.mutate.undelete_record(record)
    with pytest.raises(UnicatError):
        unicaterror.mutate.permanent_delete_record(record)


def test_mutate_assets(unicaterror):
    language = unicaterror.project.default_language
    asset = unicaterror.get_asset("<asset-1>")
    folderasset = unicaterror.get_asset("<asset-0>")
    localfilepath = "/tmp/localfile.ext"
    with pytest.raises(UnicatError):
        unicaterror.mutate.upload_asset(localfilepath, folderasset)
    with pytest.raises(UnicatError):
        unicaterror.mutate.upload_update_asset(localfilepath, asset)
    with pytest.raises(UnicatError):
        unicaterror.mutate.create_asset(folderasset, name="new folder")
    with pytest.raises(UnicatError):
        unicaterror.mutate.update_asset(
            asset,
            name="new",
            title={language: "New"},
            description={language: "New new new"},
        )
    with pytest.raises(UnicatError):
        unicaterror.mutate.update_asset(
            asset,
            other="transforms - know what you're doing!",
        )
    with pytest.raises(UnicatError):
        unicaterror.mutate.delete_asset(asset)
    with pytest.raises(UnicatError):
        unicaterror.mutate.undelete_asset(asset)
    with pytest.raises(UnicatError):
        unicaterror.mutate.permanent_delete_asset(asset)


def test_mutate_queries(unicaterror):
    query = unicaterror.get_query("<query-1>")
    with pytest.raises(UnicatError):
        unicaterror.mutate.create_query(
            type="record",
            name="new",
            q="",
            filter=["and", "", [["validation", "not_translated"]]],
        )
    with pytest.raises(UnicatError):
        unicaterror.mutate.update_query(
            query,
            name="updated",
            q="",
            filter=["and", "", [["validation", "not_translated"]]],
        )
    with pytest.raises(UnicatError):
        unicaterror.mutate.delete_query(query)


def test_backups(unicaterror):
    backup = unicaterror.project.get_backup(2)
    with pytest.raises(UnicatError):
        unicaterror.mutate.create_backup(created_by="unicat test", name="new")
    with pytest.raises(UnicatError):
        unicaterror.mutate.update_backup(backup, "updated")
    with pytest.raises(UnicatError):
        unicaterror.mutate.delete_backup(backup)
    with pytest.raises(UnicatError):
        unicaterror.mutate.delete_backups(unicaterror.project.backups[0:2])



from ..class_ import UnicatClass
from ..error import UnicatError


def create_class(unicat, *, name=None, label=None, fields=None, **kwargs):
    properties = {**kwargs}
    if name is not None:
        properties["name"] = name
    if label is not None and type(label) is dict:
        properties["label"] = label
    if fields is not None and type(fields) is not str:
        properties["fields"] = [
            field.gid if hasattr(field, "gid") else field for field in fields
        ]
    success, result = unicat.api.call("/classes/create", properties)
    if not success:
        raise UnicatError("create_class", result)
    return UnicatClass(unicat, result["class"])


def modify_class(unicat, class_, *, name=None, label=None, fields=None, **kwargs):
    properties = {**kwargs}
    if name is not None:
        properties["name"] = name
    if label is not None and type(label) is dict:
        properties["label"] = label
    if fields is not None and type(fields) is not str:
        properties["fields"] = [
            field.gid if hasattr(field, "gid") else field for field in fields
        ]
    properties["class"] = class_.gid
    success, result = unicat.api.call("/classes/modify", properties)
    if not success:
        raise UnicatError("modify_class", result)
    return UnicatClass(unicat, result["class"])


def modify_class_modify_layout(
    unicat,
    class_,
    *,
    name=None,
    root=None,
    components=None,
):
    layout_properties = {}
    if name is not None:
        layout_properties["name"] = name
    if root is not None:
        layout_properties["root"] = root
    if components is not None and type(components) is dict:
        layout_properties["components"] = components
    success, result = unicat.api.call(
        "/classes/layouts/modify",
        {**layout_properties, "class": class_.gid},
    )
    if not success:
        raise UnicatError("modify_class_modify_layout", result)
    return UnicatClass(unicat, result["class"])


def modify_class_add_field(unicat, class_, field):
    success, result = unicat.api.call(
        "/classes/fields/add",
        {"class": class_.gid, "field": field.gid},
    )
    if not success:
        raise UnicatError("modify_class_add_field", result)
    return UnicatClass(unicat, result["class"])


def modify_class_remove_field(unicat, class_, field):
    success, result = unicat.api.call(
        "/classes/fields/remove",
        {"class": class_.gid, "field": field.gid},
    )
    if not success:
        raise UnicatError("modify_class_remove_field", result)
    return UnicatClass(unicat, result["class"])


def commit_class(unicat, new_or_working_copy):
    success, result = unicat.api.call(
        "/classes/commit", {"class": new_or_working_copy.gid}
    )
    if not success:
        raise UnicatError("commit_class", result)
    if (
        result["class"] != new_or_working_copy.gid
        and new_or_working_copy.gid in unicat.api.data["classes"]
    ):
        del unicat.api.data["classes"][new_or_working_copy.gid]
    return UnicatClass(unicat, result["class"])


def save_as_new_class(unicat, working_copy):
    success, result = unicat.api.call(
        "/classes/save_as_new", {"class": working_copy.gid}
    )
    if not success:
        raise UnicatError("save_as_new_class", result)
    return UnicatClass(unicat, result["class"])


def delete_class(unicat, class_):
    success, result = unicat.api.call("/classes/delete", {"class": class_.gid})
    if not success:
        raise UnicatError("delete_class", result)
    if class_.gid in unicat.api.data["classes"]:
        del unicat.api.data["classes"][class_.gid]
    return True



from ..record import UnicatRecord
from ..error import UnicatError
from ..utils import _job


def create_record(unicat, parent, ordering):
    success, result = unicat.api.call(
        "/records/create",
        {"parent": parent.gid, "ordering": ordering},
    )
    if not success:
        raise UnicatError("create_record", result)
    return UnicatRecord(unicat, result["record"])


def set_record_definition(unicat, record, definition):
    success, result = unicat.api.call(
        "/records/definition/set",
        {"record": record.gid, "definition": definition.gid},
    )
    if not success:
        raise UnicatError("set_record_definition", result)
    return UnicatRecord(unicat, result["record"])


def extend_record_definition_add_class(unicat, record, class_):
    success, result = unicat.api.call(
        "/records/definition/classes/add",
        {"record": record.gid, "class": class_.gid},
    )
    if not success:
        raise UnicatError("extend_record_definition_add_class", result)
    return UnicatRecord(unicat, result["record"])


def extend_record_definition_add_field(unicat, record, field):
    success, result = unicat.api.call(
        "/records/definition/fields/add",
        {"record": record.gid, "field": field.gid},
    )
    if not success:
        raise UnicatError("extend_record_definition_add_field", result)
    return UnicatRecord(unicat, result["record"])


def extend_record_definition_add_fieldlist_field(unicat, record, fieldlist, field):
    success, result = unicat.api.call(
        "/records/definition/fieldlists/fields/add",
        {"record": record.gid, "fieldlist": fieldlist, "field": field.gid},
    )
    if not success:
        raise UnicatError("extend_record_definition_add_field", result)
    return UnicatRecord(unicat, result["record"])


def update_record(unicat, record, localizedfielddata):
    success, result = unicat.api.call(
        "/records/update", {"record": record.gid, "fields": localizedfielddata}
    )
    if not success:
        raise UnicatError("update_record", result)
    return UnicatRecord(unicat, result["record"], result["report.validation"])


def set_record_channels(unicat, record, channels, enabled):
    success, result = unicat.api.call(
        "/records/channels/set",
        {"record": record.gid, "channels": channels, "enabled": enabled},
    )
    if not success:
        raise UnicatError("set_record_channels", result)
    return UnicatRecord(unicat, result["record"])


def copy_record_channels_from_parent(unicat, record, channels=None):
    data = {"record": record.gid}
    if channels:
        data["channels"] = channels
    success, result = unicat.api.call("/records/channels/copy_from_parent", data)
    if not success:
        raise UnicatError("copy_record_channels_from_parent", result)
    return UnicatRecord(unicat, result["record"])


@_job
def copy_record_channels_down(unicat, record, channels=None):
    # becomes copy_record_channels_down(unicat, record, channels=None, *, return_job=False): UnicatRecord | UnicatJob
    data = {"record": record.gid}
    if channels:
        data["channels"] = channels
    success, result = unicat.api.call("/records/channels/copy_down", data)
    if not success:
        raise UnicatError("copy_record_channels_down", result)
    return result["job.token"], record  # unchanged, only descendants are changed


def copy_record_channels_up(unicat, record, channels=None):
    data = {"record": record.gid}
    if channels:
        data["channels"] = channels
    success, result = unicat.api.call("/records/channels/copy_up", data)
    if not success:
        raise UnicatError("copy_record_channels_up", result)
    return record  # unchanged, only ancestors are changed


def set_record_orderings(unicat, record, orderings):
    success, result = unicat.api.call(
        "/records/orderings/set",
        {"record": record.gid, "orderings": orderings},
    )
    if not success:
        raise UnicatError("set_record_orderings", result)
    return UnicatRecord(unicat, result["record"])


def link_record(unicat, parent, record, ordering):
    success, result = unicat.api.call(
        "/records/link",
        {"parent": parent.gid, "record": record.gid, "ordering": ordering},
    )
    if not success:
        raise UnicatError("link_record", result)
    return UnicatRecord(unicat, result["record"])


def delete_record(unicat, record):
    success, result = unicat.api.call("/records/delete", {"record": record.gid})
    if not success:
        raise UnicatError("delete_record", result)
    return UnicatRecord(unicat, result["record"])


def undelete_record(unicat, record):
    success, result = unicat.api.call("/records/undelete", {"record": record.gid})
    if not success:
        raise UnicatError("undelete_record", result)
    return UnicatRecord(unicat, result["record"])


@_job
def permanent_delete_record(unicat, record):
    # becomes permanent_delete_record(unicat, record, *, return_job=False): UnicatRecord | UnicatJob
    success, result = unicat.api.call(
        "/records/permanent_delete", {"record": record.gid}
    )
    if not success:
        raise UnicatError("permanent_delete_record", result)
    if record.gid in unicat.api.data["records"]:
        del unicat.api.data["records"][record.gid]
    return result["job.token"], UnicatRecord(unicat, result["parent"])



class UnicatDefinition:
    def __init__(self, unicat, gid):
        self._unicat = unicat
        self._data = self._unicat.definitions[gid]

    @property
    def gid(self):
        return self._data["gid"]

    @property
    def original(self):
        if not self._data["original"]:
            return None
        return UnicatDefinition(self._unicat, self._data["original"])

    @property
    def name(self):
        return self._data["name"]

    @property
    def label(self):
        return self._data["label"]

    @property
    def classes(self):
        from .class_ import UnicatClass

        return [UnicatClass(self._unicat, gid) for gid in self._data["classes"]]

    @property
    def classes_as_gids(self):
        return self._data["classes"]

    @property
    def fields(self):
        from .field import UnicatField

        return [UnicatField(self._unicat, gid) for gid in self._data["fields"]]

    @property
    def fields_as_gids(self):
        return self._data["fields"]

    @property
    def titlefield(self):
        from .field import UnicatField

        return UnicatField(self._unicat, self._data["titlefield"])

    @property
    def fieldlists(self):
        from .field import UnicatField

        return {
            key: [UnicatField(self._unicat, gid) for gid in value]
            for key, value in self._data["fieldlists"].items()
        }

    @property
    def layout(self):
        from .layout import UnicatLayout

        return UnicatLayout(self._unicat, self._data["layout"])

    @property
    def childdefinitions(self):
        return [
            UnicatDefinition(self._unicat, gid)
            for gid in self._data["childdefinitions"]
        ]

    @property
    def is_base(self):
        return self._data["is_base"]

    @property
    def is_new(self):
        return self._data["is_new"]

    @property
    def is_working_copy(self):
        return self._data["is_working_copy"]

    @property
    def is_extended(self):
        return self._data["is_extended"]

    @property
    def is_committed(self):
        return not self._data["is_new"] and not self._data["is_working_copy"]

    @property
    def all_fields(self):
        """All fields, including those from classes"""
        fields = []
        for class_ in self.classes:
            fields.extend(class_.fields)
        fields.extend(self.fields)
        return fields

    @property
    def all_base_fields(self):
        """All base fields, including those from base classes"""
        if not self.is_extended:
            fields = []
            for class_ in self.classes:
                fields.extend(class_.fields)
            fields.extend(self.fields)
            return fields
        else:
            return self.original.all_base_fields

    @property
    def all_extended_fields(self):
        """All extended fields, including those from extended classes"""
        if not self.is_extended:
            return []
        else:
            base_field_gids = [field.gid for field in self.all_base_fields]
            fields = []
            for class_ in self.classes:
                fields.extend(
                    [
                        field
                        for field in class_.fields
                        if field.gid not in base_field_gids
                    ]
                )
            fields.extend(
                [field for field in self.fields if field.gid not in base_field_gids]
            )
            return fields

    @property
    def base_fields(self):
        """Base fields, excluding those from classes"""
        if not self.is_extended:
            return self.fields
        else:
            return self.original.fields

    @property
    def extended_fields(self):
        """Extended fields, excluding those from classes"""
        if not self.is_extended:
            return []
        else:
            base_field_gids = [field.gid for field in self.base_fields]
            fields = [
                field for field in self.fields if field.gid not in base_field_gids
            ]
            return fields

    @property
    def base_classes(self):
        if not self.is_extended:
            return self.classes
        else:
            return self.original.classes

    @property
    def extended_classes(self):
        if not self.is_extended:
            return []
        else:
            base_class_gids = [class_.gid for class_ in self.base_classes]
            classes = [
                class_ for class_ in self.classes if class_.gid not in base_class_gids
            ]
            return classes



from ..field import UnicatField
from ..error import UnicatError

typecheck = type  # I use 'type' as a keyword argument


def create_field(
    unicat,
    *,
    name=None,
    type=None,
    is_localized=None,
    is_required=None,
    label=None,
    unit=None,
    initial=None,
    options=None,
):
    properties = {}
    if name is not None:
        properties["name"] = name
    if type is not None:
        properties["type"] = type
    if is_localized is not None and typecheck(is_localized) is bool:
        properties["is_localized"] = is_localized
    if is_required is not None and typecheck(is_required) is bool:
        properties["is_required"] = is_required
    if label is not None and typecheck(label) is dict:
        properties["label"] = label
    if unit is not None:
        properties["unit"] = unit
    if initial is not None and typecheck(initial) is dict:
        properties["initial"] = initial
    if options is not None and typecheck(options) is dict:
        properties["options"] = options
    success, result = unicat.api.call("/fields/create", properties)
    if not success:
        raise UnicatError("create_field", result)
    return UnicatField(unicat, result["field"])


def modify_field(
    unicat,
    field,
    *,
    name=None,
    type=None,
    is_localized=None,
    is_required=None,
    label=None,
    unit=None,
    initial=None,
    options=None,
):
    properties = {}
    if name is not None:
        properties["name"] = name
    if type is not None:
        properties["type"] = type
    if is_localized is not None and typecheck(is_localized) is bool:
        properties["is_localized"] = is_localized
    if is_required is not None and typecheck(is_required) is bool:
        properties["is_required"] = is_required
    if label is not None and typecheck(label) is dict:
        properties["label"] = label
    if unit is not None:
        properties["unit"] = unit
    if initial is not None and typecheck(initial) is dict:
        properties["initial"] = initial
    if options is not None and typecheck(options) is dict:
        properties["options"] = options
    properties["field"] = field.gid
    success, result = unicat.api.call("/fields/modify", properties)
    if not success:
        raise UnicatError("modify_field", result)
    return UnicatField(unicat, result["field"])


def commit_field(unicat, new_or_working_copy):
    success, result = unicat.api.call(
        "/fields/commit", {"field": new_or_working_copy.gid}
    )
    if not success:
        raise UnicatError("commit_field", result)
    if (
        result["field"] != new_or_working_copy.gid
        and new_or_working_copy.gid in unicat.api.data["fields"]
    ):
        del unicat.api.data["fields"][new_or_working_copy.gid]
    return UnicatField(unicat, result["field"])


def save_as_new_field(unicat, working_copy):
    success, result = unicat.api.call(
        "/fields/save_as_new", {"field": working_copy.gid}
    )
    if not success:
        raise UnicatError("save_as_new_field", result)
    return UnicatField(unicat, result["field"])


def delete_field(unicat, field):
    success, result = unicat.api.call("/fields/delete", {"field": field.gid})
    if not success:
        raise UnicatError("delete_field", result)
    if field.gid in unicat.api.data["fields"]:
        del unicat.api.data["fields"][field.gid]
    return True



import pytest
from unicat.transform import UnicatTransform, mergeTransforms, transformAsOptions


def test_transform():
    transform = UnicatTransform(name="transform-1", optimize=True, resize="fill")
    assert transform.name == "transform-1"
    assert transform.optimize
    assert transform.resize == "fill"
    assert transform.width == 1
    assert transform.height == 1


def test_transform_as_options():
    transform = UnicatTransform(name="transform-1", optimize=True, resize="fill")
    options = transformAsOptions(transform)
    assert options["name"] == "transform-1"
    assert options["optimize"] == "true"
    assert options["fill"] == "1,1"
    transform = UnicatTransform(
        name="transform-2",
        key="key",
        force=False,
        optimize=True,
        resize="fill",
        width=400,
        height=300,
        type="jpg",
        hotspot=(0.6, 0.4),
        crop=(0.8, 0.6),
        padding=(0, "auto", 0, "auto"),
        quality=75,
        background="#123abc",
        dpr=2,
    )
    options = transformAsOptions(transform)
    assert options["name"] == "transform-2"
    assert options["optimize"] == "true"
    assert options["fill"] == "400,300"
    assert options["key"] == "key"
    assert options["type"] == "jpg"
    assert options["force"] == "false"
    assert options["hs"] == "0.60000,0.40000"
    assert options["crop"] == "0.80000,0.60000"
    assert options["padding"] == "0,auto,0,auto"
    assert options["bg"] == "123abc"
    assert options["q"] == "75"
    assert options["dpr"] == "2"


def test_merge_transforms():
    transform1 = UnicatTransform(name="transform-1", optimize=True, resize="fill")
    transform2 = UnicatTransform(resize="fit", width=600, dpr=2)
    transform3 = UnicatTransform(optimize=False, height=300, hotspot=(0.6, 0.4))
    transform = mergeTransforms(transform1, transform2, transform3)
    assert transform.name == "transform-1"
    assert not transform.optimize
    assert transform.resize == "fit"
    assert transform.width == 600
    assert transform.height == 300
    assert transform.dpr == 2
    assert transform.hotspot == [0.6, 0.4]



from ..asset import UnicatAsset
from ..error import UnicatError


def upload_asset(unicat, localfilepath, folderasset):
    success, result = unicat.api.upload(
        "/assets/upload", {"parent": folderasset.gid}, localfilepath
    )
    if not success:
        raise UnicatError("upload_asset", result)
    return UnicatAsset(unicat, result["asset"])


def upload_update_asset(unicat, localfilepath, asset):
    success, result = unicat.api.upload(
        "/assets/upload_update", {"asset": asset.gid}, localfilepath
    )
    if not success:
        raise UnicatError("upload_update_asset", result)
    return UnicatAsset(unicat, result["asset"])


def create_asset(unicat, parentasset, name):
    success, result = unicat.api.call(
        "/assets/create", {"parent": parentasset.gid, "name": name}
    )
    if not success:
        raise UnicatError("create_asset", result)
    return UnicatAsset(unicat, result["asset"])


def update_asset(unicat, asset, *, name=None, title=None, description=None, **kwargs):
    properties = {**kwargs}
    if name is not None:
        properties["name"] = name
    if title is not None and type(title) is dict:
        properties["title"] = title
    if description is not None and type(description) is dict:
        properties["description"] = description
    properties["asset"] = asset.gid
    success, result = unicat.api.call("/assets/update", properties)
    if not success:
        raise UnicatError("update_asset", result)
    return UnicatAsset(unicat, result["asset"])


def delete_asset(unicat, asset):
    success, result = unicat.api.call("/assets/delete", {"asset": asset.gid})
    if not success:
        raise UnicatError("delete_asset", result)
    return UnicatAsset(unicat, result["asset"])


def undelete_asset(unicat, asset):
    success, result = unicat.api.call("/assets/undelete", {"asset": asset.gid})
    if not success:
        raise UnicatError("undelete_asset", result)
    return UnicatAsset(unicat, result["asset"])


def permanent_delete_asset(unicat, asset):
    success, result = unicat.api.call("/assets/permanent_delete", {"asset": asset.gid})
    if not success:
        raise UnicatError("permanent_delete_asset", result)
    if asset.gid in unicat.api.data["assets"]:
        del unicat.api.data["assets"][asset.gid]
    return True



from .record_field import UnicatRecordField


class UnicatRecordClassField:
    """A record class field behaves like a UnicatRecordField, but instead of a
    reference to the record, it holds a reference to its parent (class) field.
    """

    def __init__(self, unicat, parentfield, field, value):
        self._unicat = unicat
        self.parentfield = parentfield
        self.field = field
        self.value = value

    def __str__(self):
        return UnicatRecordField.__str__(self)

    def __bool__(self):
        return UnicatRecordField.__bool__(self)

    def __getattr__(self, name):
        return UnicatRecordField.__getattr__(self, name)

    def __getitem__(self, key):
        return UnicatRecordField.__getitem__(self, key)

    def _get_classfields(self, data):
        return UnicatRecordField._get_classfields(self, data)
